from __future__ import annotations

from collections.abc import Mapping
from typing import Any, cast

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    ApiError,
    AuthorizationRequiredEvent,
    BootstrapCountsReady,
    CsmSnapshotUpdated,
    NetworkRssiUpdated,
    NetworkSsidResultsUpdated,
    TableCsmChanged,
    TstatStatusUpdated,
    TstatTableInfoUpdated,
    UserConfiguredInventoryReady,
)
from elke27_lib.handlers import bus_ios, network_param, rule, tstat, user
from elke27_lib.states import PanelState, TstatState, UserState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[object] = []

    def __call__(self, evt: object, _ctx: DispatchContext) -> None:
        self.events.append(evt)


def _any_event(spy: _EmitSpy, kind: type) -> bool:
    return any(isinstance(evt, kind) for evt in spy.events)


def test_bus_ios_get_trouble_handler_paths() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = bus_ios.make_bus_ios_get_trouble_handler(state, emit, now=lambda: 1.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"bus_io_dev": {}}, make_ctx()) is False

    msg = {"bus_ios": {"get_trouble": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"bus_io_dev": {"get_trouble": {"error_code": 42}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"bus_io_dev": {"get_trouble": {"foo": "bar"}}}
    assert handler(msg, make_ctx(classification="BROADCAST")) is True
    assert cast(dict[str, Any], state.bus_io_status["get_trouble"])["foo"] == "bar"
    assert state.panel.last_message_at == 1.0


def test_rule_get_rules_handler_paths() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = rule.make_rule_get_rules_handler(state, emit, now=lambda: 2.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"rule": {}}, make_ctx()) is False

    msg = {"rule": {"get_rules": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"rule": {"get_rules": {"error_code": 9}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    msg = {"rule": {"get_rules": {"block_id": "x"}}}
    assert handler(msg, make_ctx()) is False

    msg = {"rule": {"get_rules": {"block_id": 0, "block_count": 2}}}
    assert handler(msg, make_ctx()) is True
    assert state.rules == {}
    assert state.rules_block_count == 2

    msg = {"rule": {"get_rules": {"block_id": 1, "block_count": 2, "data": "rule1"}}}
    assert handler(msg, make_ctx()) is True
    assert state.rules[1]["data"] == "rule1"
    assert state.panel.last_message_at == 2.0


def test_network_param_get_ssid_handler_paths() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = network_param.make_network_param_get_ssid_handler(state, emit, now=lambda: 3.0)

    assert handler({"nope": {}}, make_ctx()) is False

    msg = {"network": {"get_ssid": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"network": {"error_code": 9, "get_ssid": {"error_code": 9}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"network": {"get_ssid": {"ssids": ["a", {"ssid": "b"}, 4]}}}
    assert handler(msg, make_ctx()) is True
    assert state.network.ssid_scan_results == [{"ssid": "a"}, {"ssid": "b"}]
    assert _any_event(emit, NetworkSsidResultsUpdated)

    msg = {"network": {"get_ssid": {"ssid": "single"}}}
    assert handler(msg, make_ctx()) is True
    assert state.network.ssid_scan_results == [{"ssid": "single"}]

    msg = {"network": {"get_ssid": {"foo": "bar"}}}
    assert handler(msg, make_ctx()) is True
    assert state.network.ssid_scan_results == [{"foo": "bar"}]

    msg = {"network": {"get_ssid": ["x", "y"]}}
    assert handler(msg, make_ctx()) is True
    assert state.network.last_update_at == 3.0


def test_network_param_get_rssi_and_error_handler() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = network_param.make_network_param_get_rssi_handler(state, emit, now=lambda: 4.0)

    assert handler({"nope": {}}, make_ctx()) is False

    msg = {"network": {"get_rssi": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"network": {"error_code": 8, "get_rssi": {"error_code": 8}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"network": {"get_rssi": {"rssi": "12"}}}
    assert handler(msg, make_ctx()) is True
    assert state.network.rssi == 12
    assert _any_event(emit, NetworkRssiUpdated)

    msg = {"network": {"rssi": 5}}
    assert handler(msg, make_ctx()) is True
    assert state.network.rssi == 5

    error_handler = network_param.make_network_error_handler(state, emit, lambda: 0.0)
    assert error_handler({"nope": {}}, make_ctx()) is False
    assert (
        error_handler({"network": {"error_code": "11008", "error_message": "no"}}, make_ctx())
        is True
    )
    assert _any_event(emit, AuthorizationRequiredEvent)
    emit.events.clear()
    assert error_handler({"network": {"error_code": 7, "error_message": "bad"}}, make_ctx()) is True
    assert _any_event(emit, ApiError)
    assert error_handler({"network": {"error_code": "notint"}}, make_ctx()) is False


def test_network_param_helpers() -> None:
    net_obj: Mapping[str, object] = {"get_ssid": ["foo", {"ssid": "bar"}, 1]}
    assert network_param._normalize_ssid_results(None, net_obj) == []
    assert network_param._normalize_ssid_results(["a", {"ssid": "b"}], net_obj) == [
        {"ssid": "a"},
        {"ssid": "b"},
    ]
    assert network_param._normalize_ssid_results("solo", net_obj) == [{"ssid": "solo"}]
    assert network_param._normalize_ssid_results({"ssid": "named"}, net_obj) == [{"ssid": "named"}]
    assert network_param._normalize_ssid_results({"foo": "bar"}, net_obj) == [{"foo": "bar"}]
    assert network_param._normalize_ssid_results({"scan": ["x", "y"]}, net_obj) == [
        {"ssid": "x"},
        {"ssid": "y"},
    ]
    assert network_param._normalize_ssid_results(123, net_obj) == [{"ssid": "foo"}, {"ssid": "bar"}]
    assert network_param._normalize_ssid_results(None, {"get_ssid": "bad"}) == []
    assert network_param._normalize_ssid_results(123, {"get_ssid": "bad"}) == []

    assert network_param._extract_rssi({"rssi": 7}, {}) == 7
    assert network_param._extract_rssi({"rssi": "9"}, {}) == 9
    assert network_param._extract_rssi({"rssi": "bad"}, {}) is None
    assert network_param._extract_rssi({"rssi": {}}, {}) is None
    assert network_param._extract_rssi(None, {"rssi": 11}) == 11


def test_tstat_get_status_and_table_info() -> None:
    state = PanelState()
    emit = _EmitSpy()

    status_handler = tstat.make_tstat_get_status_handler(state, emit, now=lambda: 5.0)
    assert status_handler({"nope": {}}, make_ctx()) is False
    assert status_handler({"tstat": {}}, make_ctx()) is False

    msg = {"tstat": {"get_status": {"error_code": 10}}}
    assert status_handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    msg = {"tstat": {"get_status": {"tstat_id": 0}}}
    assert status_handler(msg, make_ctx()) is False

    emit.events.clear()
    msg = {
        "tstat": {
            "get_status": {
                "tstat_id": 1,
                "temperature": 70,
                "fan_mode": "AUTO",
                "battery level": 99,
                "prec": [1, 2],
                "custom": "x",
            }
        }
    }
    assert status_handler(msg, make_ctx()) is True
    assert state.tstats[1].temperature == 70
    assert state.tstats[1].battery_level == 99
    assert state.tstats[1].prec == [1, 2]
    assert state.tstats[1].fields["custom"] == "x"
    assert _any_event(emit, TstatStatusUpdated)

    table_handler = tstat.make_tstat_get_table_info_handler(state, emit, now=lambda: 6.0)
    assert table_handler({"nope": {}}, make_ctx()) is False
    assert table_handler({"tstat": {}}, make_ctx()) is False

    msg = {"tstat": {"get_table_info": {"error_code": 8}}}
    assert table_handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    state.table_info_known.update({"area", "zone", "output"})
    msg = {"tstat": {"table_info": {"table_elements": 2, "table_csm": 4}}}
    assert table_handler(msg, make_ctx()) is True
    assert state.table_info_by_domain["tstat"]["table_elements"] == 2
    assert _any_event(emit, TableCsmChanged)
    assert _any_event(emit, TstatTableInfoUpdated)
    assert _any_event(emit, CsmSnapshotUpdated)
    assert _any_event(emit, BootstrapCountsReady)


def test_tstat_helpers() -> None:
    tstat_state = TstatState(tstat_id=1)
    changed: set[str] = set()
    tstat._apply_tstat_status_fields(
        tstat_state,
        {
            "temperature": 70,
            "cool_setpoint": 68,
            "heat_setpoint": 72,
            "mode": "AUTO",
            "fan_mode": "ON",
            "humidity": 50,
            "rssi": -40,
            "battery_level": 80,
            "prec": [1, "bad"],
            "extra": "x",
        },
        changed,
    )
    assert tstat_state.temperature == 70
    assert tstat_state.cool_setpoint == 68
    assert tstat_state.heat_setpoint == 72
    assert tstat_state.fan_mode == "ON"
    assert tstat_state.humidity == 50
    assert tstat_state.rssi == -40
    assert tstat_state.battery_level == 80
    assert tstat_state.prec is None
    assert tstat_state.fields["extra"] == "x"

    assert tstat._extract_table_csm({"table_csm": True}, domain="tstat") is None
    assert tstat._extract_table_csm({"table_csm": 5}, domain="tstat") == 5
    assert tstat._extract_table_csm({"table_csm": 3.0}, domain="tstat") == 3
    assert tstat._extract_table_csm({"table_csm": "4"}, domain="tstat") == 4
    assert tstat._extract_table_csm({"table_csm": "bad"}, domain="tstat") is None
    assert tstat._extract_table_csm({"other": 1}, domain="tstat") is None


def test_user_handlers() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = user.make_user_get_configured_handler(state, emit, now=lambda: 7.0)

    assert handler({"nope": {}}, make_ctx()) is False
    assert handler({"user": {}}, make_ctx()) is False

    msg = {"user": {"get_configured": {"error_code": 11008}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"user": {"get_configured": {"error_code": 9}}}
    assert handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    emit.events.clear()
    msg = {"user": {"get_configured": {"users": [1, 2], "block_id": 1, "block_count": 1}}}
    assert handler(msg, make_ctx()) is True
    assert state.inventory.configured_users == {1, 2}
    assert state.inventory.configured_users_complete is True
    assert _any_event(emit, UserConfiguredInventoryReady)
    assert state.panel.last_message_at == 7.0

    attribs_handler = user.make_user_get_attribs_handler(state, emit, now=lambda: 8.0)
    assert attribs_handler({"nope": {}}, make_ctx()) is False
    assert attribs_handler({"user": {}}, make_ctx()) is False

    msg = {"user": {"get_attribs": {"error_code": 11008, "user_id": 1}}}
    assert attribs_handler(msg, make_ctx()) is True
    assert _any_event(emit, AuthorizationRequiredEvent)

    emit.events.clear()
    msg = {"user": {"get_attribs": {"error_code": 3, "user_id": 1}}}
    assert attribs_handler(msg, make_ctx()) is True
    assert _any_event(emit, ApiError)

    msg = {"user": {"get_attribs": {"user_id": 0}}}
    assert attribs_handler(msg, make_ctx()) is False

    msg = {
        "user": {
            "get_attribs": {
                "user_id": 1,
                "name": "  Alice ",
                "group_id": 2,
                "enabled": True,
                "pin": 1234,
                "flags": [1],
                "extra": "x",
            }
        }
    }
    assert attribs_handler(msg, make_ctx()) is True
    user_state = state.users[1]
    assert user_state.name == "Alice"
    assert user_state.group_id == 2
    assert user_state.enabled is True
    assert user_state.pin == 1234
    assert user_state.flags == [1]
    assert user_state.fields["extra"] == "x"
    assert state.panel.last_message_at == 8.0


def test_user_helpers() -> None:
    assert user._extract_configured_ids({"users": [1, 0, 2, "x"]}) == {1, 2}
    assert user._extract_configured_ids({"user_ids": [3]}) == {3}
    assert user._extract_configured_ids({"configured_users": [4]}) == {4}
    assert user._extract_configured_ids({"configured_user_ids": [5]}) == {5}
    assert user._extract_configured_ids({"none": []}) == set()
    assert user._normalize_name(None) is None
    assert user._normalize_name("   ") is None
    assert user._normalize_name(" Bob ") == "Bob"

    user_state = UserState(user_id=1)
    changed: set[str] = set()
    user._apply_user_attribs(
        user_state,
        {"name": " Alice ", "group_id": 2, "enabled": True, "pin": 1234, "flags": [1], "x": 1},
        changed,
    )
    assert user_state.name == "Alice"
    assert user_state.group_id == 2
    assert user_state.enabled is True
    assert user_state.pin == 1234
    assert user_state.flags == [1]
    assert user_state.fields["x"] == 1



# test/conftest.py
from __future__ import annotations

import contextlib
import os
import pathlib
import uuid
from collections.abc import AsyncIterator, Generator
from dataclasses import dataclass
from typing import Any, cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.linking import E27Identity
from test.helpers.internal import get_kernel, set_private
from test.helpers.reporter import Reporter


@dataclass(frozen=True)
class LiveCredentials:
    access_code: str
    passphrase: str


def get_env(name: str) -> str | None:
    value = os.environ.get(name)
    if value is None or value.strip() == "":
        return None
    return value


@pytest.fixture(scope="function")
async def live_e27_client(request: pytest.FixtureRequest) -> AsyncIterator[Elke27Client]:
    if get_env("ELKE27_LIVE") != "1":
        pytest.skip("ELKE27_LIVE not set; skipping live E27 tests.")

    host = get_env("ELKE27_HOST")
    access_code = get_env("ELKE27_ACCESS_CODE")
    passphrase = get_env("ELKE27_PASSPHRASE")
    if not host or not access_code or not passphrase:
        pytest.skip("Missing E27 live env vars; source ~/elk-e27-env-vars.sh")

    port = int(get_env("ELKE27_PORT") or "2101")

    client = Elke27Client()
    identity = E27Identity(
        mn=get_env("ELKE27_MN") or "CODEx",
        sn=get_env("ELKE27_SN") or "LIVE",
        fwver=get_env("ELKE27_FWVER") or "0",
        hwver=get_env("ELKE27_HWVER") or "0",
        osver=get_env("ELKE27_OSVER") or "0",
    )
    creds = LiveCredentials(access_code=access_code, passphrase=passphrase)

    skip_bootstrap = any("test_live_e27_keepalive.py" in arg for arg in request.config.args)
    if skip_bootstrap:
        kernel = get_kernel(client)

        def _noop_bootstrap_requests() -> None:
            return None

        set_private(kernel, "_bootstrap_requests", _noop_bootstrap_requests)

    try:
        link_keys = await client.async_link(
            host=host,
            port=port,
            access_code=creds.access_code,
            passphrase=creds.passphrase,
            client_identity=identity,
        )
    except Exception as exc:  # noqa: BLE001
        pytest.fail(f"E27 link failed; check credentials and panel availability: {exc}")

    try:
        await client.async_connect(host, port, link_keys)
    except Exception as exc:  # noqa: BLE001
        pytest.fail(f"E27 connect failed; check panel availability: {exc}")

    if not skip_bootstrap:
        await client.wait_ready(timeout_s=10.0)

    try:
        yield client
    finally:
        await client.async_disconnect()


def _get_or_create_run_id(cfg: pytest.Config) -> str:
    """Return a stable run_id for the entire pytest invocation.

    A single pytest invocation should write to a single jsonl file. If run_id is
    generated in more than one place (e.g., hooks + fixtures), you'll end up with
    multiple output files for what you think is "one run".
    """
    cfg_any = cast(Any, cfg)
    run_id = getattr(cfg_any, "_e27_run_id", None)
    if not run_id:
        run_id = uuid.uuid4().hex
        cfg_any._e27_run_id = run_id
    return run_id


def _get_or_create_report_path(cfg: pytest.Config) -> pathlib.Path:
    """Return the jsonl path for this pytest invocation (stable)."""
    cfg_any = cast(Any, cfg)
    p = getattr(cfg_any, "_e27_report_path", None)
    if p is None:
        base_dir = pathlib.Path(cfg.getoption("--e27-artifacts-dir"))
        #        artifacts_dir = base_dir / "test_runs"
        artifacts_dir = base_dir
        artifacts_dir.mkdir(parents=True, exist_ok=True)
        p = artifacts_dir / f"{_get_or_create_run_id(cfg)}.jsonl"
        cfg_any._e27_report_path = p
    return p


def pytest_configure(config: pytest.Config) -> None:
    """Initialize run_id/report path early so hooks/fixtures share them."""
    _get_or_create_report_path(config)


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("e27")
    group.addoption(
        "--e27-live",
        action="store_true",
        default=False,
        help="Enable tests marked as live (requires real panel + credentials).",
    )
    group.addoption(
        "--e27-report",
        action="store",
        default=os.getenv("ELK_E27_REPORT_FORMAT", "jsonl"),
        choices=("jsonl", "yaml", "both", "none"),
        help="Test artifact format: jsonl (default), yaml, both, or none. "
        "YAML is derived from JSON records.",
    )
    group.addoption(
        "--e27-artifacts-dir",
        action="store",
        default=os.getenv("ELK_E27_ARTIFACTS_DIR", "artifacts/test_runs"),
        help="Directory to write E27 test artifacts (default: artifacts/test_runs).",
    )


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_runtest_makereport(
    item: pytest.Item, call: pytest.CallInfo[object]
) -> Generator[None, Any, Any]:
    # Attach the report to the item so hooks/fixtures can access pass/fail and exceptions.
    del call
    outcome = yield
    rep = outcome.get_result()
    setattr(item, f"rep_{rep.when}", rep)
    if rep.when == "teardown" and rep.failed:
        mark = item.get_closest_marker("xfail")
        if mark is not None:
            reason = mark.kwargs.get("reason")
            if reason is None and mark.args:
                reason = str(mark.args[0])
            if reason and "teardown" in str(reason).lower():
                rep.outcome = "skipped"
                rep.wasxfail = reason


@pytest.fixture()
def fail_teardown(request: pytest.FixtureRequest) -> Generator[None, None, None]:
    def fin():
        pytest.fail("intentional teardown failure", pytrace=True)

    request.addfinalizer(fin)
    yield


@pytest.fixture(scope="session")
def e27_run_id(request: pytest.FixtureRequest) -> str:
    # Stable per pytest invocation (shared with hooks).
    return _get_or_create_run_id(request.config)


@pytest.fixture()
def reporter(request: pytest.FixtureRequest, e27_run_id: str) -> Generator[Reporter, None, None]:
    """
    Create a per-test Reporter and attach it to the item.
    We do NOT finalize here; finalization happens in pytest_runtest_teardown,
    after rep_teardown has been generated by pytest.
    """
    cfg = request.config
    report_mode = str(cfg.getoption("--e27-report")).lower()
    artifacts_dir = pathlib.Path(str(cfg.getoption("--e27-artifacts-dir")))
    emit_jsonl = report_mode in ("jsonl", "both")
    emit_yaml = report_mode in ("yaml", "both")
    enable = report_mode != "none"

    node = cast(pytest.Item, request.node)
    node_any = cast(Any, node)
    r = Reporter(
        run_id=e27_run_id,
        test_id=node.nodeid,
        artifacts_dir=artifacts_dir,
        emit_jsonl=emit_jsonl,
        emit_yaml=emit_yaml,
        enable=enable,
    )
    r.test_start()

    # Make it accessible to hooks
    node_any._e27_reporter = r

    yield r


def _finalize_reporter_for_item(item: pytest.Item) -> None:
    item_any = cast(Any, item)
    r: Reporter | None = getattr(item_any, "_e27_reporter", None)
    if r is None:
        return

    rep_setup = getattr(item, "rep_setup", None)
    rep_call = getattr(item, "rep_call", None)
    rep_teardown = getattr(item, "rep_teardown", None)

    outcome = "unknown"
    longrepr = None
    when = None

    # Setup failures always win
    if rep_setup is not None and rep_setup.failed:
        outcome = "fail"
        longrepr = rep_setup.longrepr
        when = "setup"

    # Call phase
    elif rep_call is not None:
        when = "call"
        if rep_call.passed:
            outcome = "pass"
        elif rep_call.skipped:
            outcome = "skip"
            longrepr = rep_call.longrepr
        else:
            outcome = "fail"
            longrepr = rep_call.longrepr

    # Teardown failures override a pass
    if outcome == "pass" and rep_teardown is not None and rep_teardown.failed:
        outcome = "fail"
        longrepr = rep_teardown.longrepr
        when = "teardown"

    r.test_end(outcome=outcome, when=when, longrepr=longrepr)
    r.finalize()

    # Prevent double-finalization if pytest calls teardown hooks unusually
    with contextlib.suppress(Exception):
        delattr(item_any, "_e27_reporter")


def pytest_runtest_teardown(item: pytest.Item, nextitem: pytest.Item | None) -> None:
    """
    Finalize after teardown completes and rep_teardown is available.
    This is the key change for reliable teardown failure reporting.
    """
    del nextitem
    _finalize_reporter_for_item(item)


def pytest_runtest_setup(item: pytest.Item) -> None:
    """
    If setup itself fails before the test body can request the reporter fixture,
    we still want the reporter to exist and capture the failure.
    So: create a reporter eagerly if reporting is enabled.
    """
    cfg = item.config
    report_mode = str(cfg.getoption("--e27-report")).lower()
    if report_mode == "none":
        return

    item_any = cast(Any, item)
    if hasattr(item_any, "_e27_reporter"):
        return

    cfg_any = cast(Any, cfg)
    run_id = getattr(cfg_any, "_e27_run_id", None)
    if run_id is None:
        run_id = uuid.uuid4().hex
        cfg_any._e27_run_id = run_id

    artifacts_dir = pathlib.Path(str(cfg.getoption("--e27-artifacts-dir")))
    emit_jsonl = report_mode in ("jsonl", "both")
    emit_yaml = report_mode in ("yaml", "both")

    r = Reporter(
        run_id=run_id,
        test_id=item.nodeid,
        artifacts_dir=artifacts_dir,
        emit_jsonl=emit_jsonl,
        emit_yaml=emit_yaml,
        enable=True,
    )
    r.test_start()
    item_any._e27_reporter = r


@pytest.fixture()
def fail_setup():
    raise AssertionError("intentional setup failure")



"""Constants used across package."""

from enum import IntEnum

REDACT_DIAGNOSTICS = True


class E27ErrorCode(IntEnum):
    ELKERR_NONE = 0

    ELKERR_UNRECOGNIZED_CMD = 11001
    ELKERR_ID_EXISTS = 11002
    ELKERR_INVALID_TABLE = 11003
    ELKERR_INVALID_PARAM = 11004
    ELKERR_NOT_ENOUGH_SPACE = 11005
    ELKERR_INVALID_ID = 11006
    ELKERR_TIMEOUT = 11007
    ELKERR_NOAUTH = 11008
    ELKERR_MISSING = 11009
    ELKERR_CALLBACK_INVALID = 11010
    ELKERR_NO_RESP = 11011
    ELKERR_DUPLICATE_PIN = 11012
    ELKERR_DUPLICATE_NAME = 11013
    ELKERR_INVALID_MEDIA = 11014
    ELKERR_NOT_READY = 11015
    ELKERR_IN_USE = 11016
    ELKERR_STORAGE_NOT_READY = 11017
    ELKERR_UNKNOWN_BUS_MODULE = 11018
    ELKERR_IN_USE_BY_RULE = 11019
    ELKERR_STRING_TRUNCATED = 11020
    ELKERR_INVALID_CACHE = 11021
    ELKERR_INVALID_MODE = 11022
    ELKERR_NOT_BYPASSABLE = 11023
    ELKERR_REBOOT_REQ = 11024
    ELKERR_CLOUD_BROKER_ALREADY_CONNECTED = 11025
    ELKERR_CLOUD_TOO_MANY_SESSIONS = 11026
    ELKERR_IN_ALARM = 11027
    ELKERR_NOT_ALLOWED_WHEN_ARMED = 11028
    ELKERR_2WSMK_NOT_ALLOWED = 11029
    ELKERR_INCOMPATIBLE = 11030
    ELKERR_UPDATING = 11031
    ELKERR_DUPLICATE_SN = 11032
    ELKERR_INVALID_AREA = 11033
    ELKERR_NOT_ENABLE = 11034
    ELKERR_RCVD_DONE = 11035
    ELKERR_INVALID_SESSION = 11036
    ELKERR_INVALID_PIN = 11037
    ELKERR_DURESS_USER_NOT_ALLOWED = 11038
    ELKERR_ZWAVE_BUSY = 11039
    ELKERR_ZWAVE_DONE = 11040

    ELKERR_UNKNOWN = 0xFFFFFFFF



"""Barrier domain request generators."""

from __future__ import annotations

ResponseKey = tuple[str, str]


def generator_barrier_get_table_info() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("barrier", "get_table_info")


def generator_barrier_get_configured(*, block_id: int = 1) -> tuple[dict[str, object], ResponseKey]:
    if block_id < 1:
        raise ValueError(f"block_id must be an int >= 1 (got {block_id!r})")
    return {"block_id": block_id}, ("barrier", "get_configured")


def generator_barrier_get_attribs(*, barrier_id: int) -> tuple[dict[str, object], ResponseKey]:
    if barrier_id < 1:
        raise ValueError(f"barrier_id must be an int >= 1 (got {barrier_id!r})")
    return {"barrier_id": barrier_id}, ("barrier", "get_attribs")


def generator_barrier_get_status(*, barrier_id: int) -> tuple[dict[str, object], ResponseKey]:
    if barrier_id < 1:
        raise ValueError(f"barrier_id must be an int >= 1 (got {barrier_id!r})")
    return {"barrier_id": barrier_id}, ("barrier", "get_status")


def generator_barrier_set_status(
    *, barrier_id: int, status: str
) -> tuple[dict[str, object], ResponseKey]:
    if barrier_id < 1:
        raise ValueError(f"barrier_id must be an int >= 1 (got {barrier_id!r})")
    normalized = status.strip().upper()
    if normalized not in {"OPEN", "CLOSE", "STOP"}:
        raise ValueError(f"status must be OPEN/CLOSE/STOP (got {status!r})")
    return {"barrier_id": barrier_id, "status": normalized}, ("barrier", "set_status")



from __future__ import annotations

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    CsmSnapshotUpdated,
    DomainCsmChanged,
    Event,
    TableCsmChanged,
    ZoneTableInfoUpdated,
)
from elke27_lib.handlers.control import make_control_authenticate_handler
from elke27_lib.handlers.zone import make_zone_get_table_info_handler
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[Event] = []

    def __call__(self, evt: Event, _ctx: DispatchContext) -> None:
        self.events.append(evt)


_Ctx = make_ctx


def test_authenticate_updates_domain_csm_and_snapshot() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_control_authenticate_handler(state, emit, now=lambda: 123.0)

    msg = {
        "control": {
            "authenticate": {
                "error_code": 0,
                "zone_csm": 10,
                "area_csm": "11",
            }
        }
    }
    assert handler(msg, _Ctx()) is True
    assert state.domain_csm_by_name == {"zone": 10, "area": 11}
    assert state.csm_snapshot is not None
    assert state.csm_snapshot.version == 1

    domain_events = [evt for evt in emit.events if isinstance(evt, DomainCsmChanged)]
    snapshot_events = [evt for evt in emit.events if isinstance(evt, CsmSnapshotUpdated)]

    assert {evt.domain for evt in domain_events} == {"zone", "area"}
    assert len(snapshot_events) == 1


def test_table_info_csm_changes_emit_snapshot_once() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_zone_get_table_info_handler(state, emit, now=lambda: 123.0)

    msg = {
        "zone": {
            "get_table_info": {
                "table_elements": 5,
                "increment_size": 1,
                "table_csm": 7,
            }
        }
    }
    assert handler(msg, _Ctx()) is True
    assert state.table_csm_by_domain["zone"] == 7
    assert state.csm_snapshot is not None
    assert state.csm_snapshot.version == 1

    table_events = [evt for evt in emit.events if isinstance(evt, TableCsmChanged)]
    snapshot_events = [evt for evt in emit.events if isinstance(evt, CsmSnapshotUpdated)]
    table_info_events = [evt for evt in emit.events if isinstance(evt, ZoneTableInfoUpdated)]
    assert len(table_events) == 1
    assert len(snapshot_events) == 1
    assert table_info_events[0].table_csm == 7

    emit.events.clear()
    assert handler(msg, _Ctx()) is True
    table_events = [evt for evt in emit.events if isinstance(evt, TableCsmChanged)]
    snapshot_events = [evt for evt in emit.events if isinstance(evt, CsmSnapshotUpdated)]
    table_info_events = [evt for evt in emit.events if isinstance(evt, ZoneTableInfoUpdated)]
    assert len(table_events) == 0
    assert len(snapshot_events) == 0
    assert table_info_events[0].table_csm == 7



"""Lock domain request generators."""

from __future__ import annotations

ResponseKey = tuple[str, str]


def generator_lock_get_table_info() -> tuple[dict[str, object], ResponseKey]:
    return {}, ("lock", "get_table_info")


def generator_lock_get_configured(*, block_id: int = 1) -> tuple[dict[str, object], ResponseKey]:
    if block_id < 1:
        raise ValueError(f"block_id must be an int >= 1 (got {block_id!r})")
    return {"block_id": block_id}, ("lock", "get_configured")


def generator_lock_get_attribs(*, lock_id: int) -> tuple[dict[str, object], ResponseKey]:
    if lock_id < 1:
        raise ValueError(f"lock_id must be an int >= 1 (got {lock_id!r})")
    return {"lock_id": lock_id}, ("lock", "get_attribs")


def generator_lock_get_status(*, lock_id: int) -> tuple[dict[str, object], ResponseKey]:
    if lock_id < 1:
        raise ValueError(f"lock_id must be an int >= 1 (got {lock_id!r})")
    return {"lock_id": lock_id}, ("lock", "get_status")


def generator_lock_set_status(
    *, lock_id: int, status: str
) -> tuple[dict[str, object], ResponseKey]:
    if lock_id < 1:
        raise ValueError(f"lock_id must be an int >= 1 (got {lock_id!r})")
    normalized = status.strip().upper()
    if normalized not in {"ON", "OFF"}:
        raise ValueError(f"status must be ON/OFF (got {status!r})")
    return {"lock_id": lock_id, "status": normalized}, ("lock", "set_status")



from __future__ import annotations

from collections.abc import Callable
from typing import Any, cast

from elke27_lib.handlers import zone as zone_handlers
from elke27_lib.states import PanelState
from test.helpers.internal import get_private

_RECONCILE_BULK_ZONE_STATUS = cast(
    Callable[..., object], get_private(zone_handlers, "_reconcile_bulk_zone_status")
)


def test_zone_status_string_updates_zone_state() -> None:
    state = PanelState()
    state.get_or_create_zone(1)
    state.get_or_create_zone(2)
    state.get_or_create_zone(3)
    payload = {"status": "1A4"}

    outcome = _RECONCILE_BULK_ZONE_STATUS(state, payload, now=1.0)

    updated_ids = cast(Any, outcome).updated_ids
    assert updated_ids == (1, 2, 3)

    z1 = state.zones[1]
    assert z1.status_code == "1"
    assert z1.enabled is True
    assert z1.violated is False
    assert z1.trouble is False
    assert z1.bypassed is False

    z2 = state.zones[2]
    assert z2.status_code == "A"
    assert z2.enabled is True
    assert z2.violated is True
    assert z2.trouble is False
    assert z2.bypassed is False

    z3 = state.zones[3]
    assert z3.status_code == "4"
    assert z3.enabled is False
    assert z3.violated is False
    assert z3.trouble is False
    assert z3.bypassed is False



# test/test_e27_session.py
#
# Unit tests for session.py (DDR-0034 / Option A)
#
# These tests validate Session as a transport + crypto + framing boundary:
# - TCP connect lifecycle + HELLO handshake + key storage
# - framing pump uses DeframeState + deframe_feed(state, chunk)
# - send_json encrypts schema-0 and frames via frame_build, then sendall
# - recv_json deframes + decrypts schema-0 + parses JSON dict
# - pump_once dispatches on_message, handles timeout, and disconnects on errors
#
# IMPORTANT:
# - Session is NOT expected to perform API_LINK, authenticate, or workflows.
# - These tests use fakes/mocks for socket, perform_hello, framing, and crypto helpers.

from __future__ import annotations

import socket
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, cast

import pytest

# Adjust import path if your session module lives elsewhere.
# The tests are written to be resilient by monkeypatching module-level dependencies.
from elke27_lib import linking
from elke27_lib import session as session_mod
from elke27_lib.framing import DeframeState
from test.helpers.internal import get_private, set_private


@dataclass
class _HelloKeys:
    session_id: int
    session_key_hex: str
    hmac_key_hex: str


class _FakeSocket:
    """
    Minimal socket fake used by Session.connect() / _recv_some() / _send_all().
    """

    def __init__(self) -> None:
        self.connected_to: tuple[str, int] | None = None
        self.timeouts: list[float] = []
        self.closed: bool = False
        self.sent: list[bytes] = []
        self._recv_queue: list[bytes] = []

    def settimeout(self, t: float) -> None:
        self.timeouts.append(t)

    def connect(self, addr: tuple[str, int]) -> None:
        self.connected_to = addr

    def close(self) -> None:
        self.closed = True

    def sendall(self, data: bytes) -> None:
        self.sent.append(data)

    def recv(self, _max_bytes: int) -> bytes:
        # Session._recv_some treats b"" as socket closed.
        if not self._recv_queue:
            return b""
        return self._recv_queue.pop(0)

    def queue_recv(self, *chunks: bytes) -> None:
        self._recv_queue.extend(chunks)


@dataclass
class _DeframeResult:
    ok: bool
    frame_no_crc: bytes = b""


@dataclass
class _DecryptEnvelope:
    payload: bytes


def _make_session_ready(monkeypatch: pytest.MonkeyPatch) -> tuple[session_mod.Session, _FakeSocket]:
    """
    Create a Session instance that is already connected/ready without running connect().
    """
    _ = monkeypatch
    cfg = session_mod.SessionConfig(host="127.0.0.1", port=2101, keepalive_enabled=False)
    identity = linking.E27Identity(
        mn="0222", sn="001122334455", fwver="1.0", hwver="1.0", osver="1.0"
    )
    s = session_mod.Session(cfg, client_identity=identity, link_key_hex="00" * 16)

    fake_sock = _FakeSocket()

    # Mark session as "ready"
    s.sock = cast(socket.socket, cast(object, fake_sock))
    set_private(s, "_deframe_state", DeframeState())
    s.info = session_mod.SessionInfo(
        session_id=123, session_key_hex="11" * 16, session_hmac_hex="22" * 20
    )
    # Session methods under test require the ACTIVE state. In production this is set by connect().
    # For these unit tests we intentionally bypass connect() and set state directly.
    s.state = session_mod.SessionState.ACTIVE

    return s, fake_sock


def test_connect_establishes_tcp_and_performs_hello(monkeypatch: pytest.MonkeyPatch) -> None:
    fake_sock = _FakeSocket()

    # Patch socket.socket(...) to return our fake.
    def _fake_socket_ctor(*_args: Any, **_kwargs: Any) -> _FakeSocket:
        return fake_sock

    monkeypatch.setattr(socket, "socket", _fake_socket_ctor)

    # Patch HELLO to return deterministic keys.
    keys = _HelloKeys(session_id=7, session_key_hex="aa" * 16, hmac_key_hex="bb" * 20)

    def _fake_perform_hello(
        *,
        sock: Any,
        client_identity: linking.E27Identity,
        linkkey_hex: str,
        timeout_s: float,
    ) -> _HelloKeys:
        _ = client_identity
        assert sock is fake_sock
        assert identity.mn == "0222"
        assert linkkey_hex == "cc" * 16
        assert timeout_s == 9.0
        return keys

    monkeypatch.setattr(session_mod, "perform_hello", _fake_perform_hello)

    cfg = session_mod.SessionConfig(
        host="10.0.0.5",
        port=2101,
        connect_timeout_s=3.0,
        io_timeout_s=0.25,
        hello_timeout_s=9.0,
        keepalive_enabled=False,
    )
    identity = linking.E27Identity(
        mn="0222", sn="001122334455", fwver="1.0", hwver="1.0", osver="1.0"
    )
    s = session_mod.Session(cfg, client_identity=identity, link_key_hex="cc" * 16)

    captured: dict[str, Any] = {}

    def _on_connected(info: session_mod.SessionInfo) -> None:
        captured["info"] = info

    s.on_connected = _on_connected

    info = s.connect()

    # TCP connect and timeouts were set correctly.
    assert fake_sock.connected_to == ("10.0.0.5", 2101)
    assert fake_sock.timeouts[0] == 3.0  # connect timeout
    assert fake_sock.timeouts[1] == 0.25  # io/pump cadence timeout

    # HELLO result stored as SessionInfo and returned.
    assert isinstance(info, session_mod.SessionInfo)
    assert info.session_id == 7
    assert info.session_key_hex == "aa" * 16
    assert info.session_hmac_hex == "bb" * 20


def test_close_is_idempotent(monkeypatch: pytest.MonkeyPatch) -> None:
    _ = monkeypatch
    s, fake_sock = _make_session_ready(monkeypatch)

    s.close()
    assert fake_sock.closed is True
    assert s.sock is None
    assert get_private(s, "_deframe_state") is None
    assert s.info is None

    # Calling again should not raise.
    s.close()
    assert s.sock is None


def test_send_json_encrypts_frames_and_sendall(monkeypatch: pytest.MonkeyPatch) -> None:
    s, fake_sock = _make_session_ready(monkeypatch)

    # Patch encrypt_schema0_envelope to return a protocol byte + ciphertext.
    seen_seq: list[int] = []

    def _fake_encrypt_schema0_envelope(
        *,
        payload: bytes,
        session_key: bytes,
        src: int,
        dest: int,
        head: int,
        envelope_seq: int,
        _iv: bytes = b"",
    ) -> tuple[int, bytes]:
        assert s.info is not None
        assert session_key == bytes.fromhex(s.info.session_key_hex)
        # Ensure JSON is compact separators (",", ":") and utf-8.
        decoded = payload.decode("utf-8")
        assert decoded == '{"a":1,"b":"x"}'
        assert src == 1
        assert dest == 0
        assert head == 0
        seen_seq.append(envelope_seq)
        return 0x83, b"CIPHERTEXT"

    monkeypatch.setattr(session_mod, "encrypt_schema0_envelope", _fake_encrypt_schema0_envelope)

    # Patch frame_build to create deterministic framed bytes.
    def _fake_frame_build(*, protocol_byte: int, data_frame: bytes) -> bytes:
        assert protocol_byte == 0x83
        assert data_frame == b"CIPHERTEXT"
        return b"FRAMED:" + bytes([protocol_byte]) + data_frame

    monkeypatch.setattr(session_mod, "frame_build", _fake_frame_build)

    s.send_json({"a": 1, "b": "x"})
    s.send_json({"a": 1, "b": "x"})
    assert fake_sock.sent == [
        b"FRAMED:" + b"\x83" + b"CIPHERTEXT",
        b"FRAMED:" + b"\x83" + b"CIPHERTEXT",
    ]
    assert seen_seq == [1, 2]


def test_envelope_seq_wraps(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)
    set_private(s, "_tx_envelope_seq", 2147483647)
    captured: list[int] = []

    def _fake_encrypt_schema0_envelope(
        *,
        payload: bytes,
        session_key: bytes,
        src: int,
        dest: int,
        head: int,
        envelope_seq: int,
        iv: bytes = b"",
    ) -> tuple[int, bytes]:
        _ = payload, session_key, src, dest, head, iv
        captured.append(envelope_seq)
        return 0x80, b"X"

    monkeypatch.setattr(session_mod, "encrypt_schema0_envelope", _fake_encrypt_schema0_envelope)

    def _fake_frame_build(*, protocol_byte: int, data_frame: bytes) -> bytes:
        _ = protocol_byte, data_frame
        return b"X"

    monkeypatch.setattr(session_mod, "frame_build", _fake_frame_build)

    s.send_json({"x": 1})
    s.send_json({"x": 1})
    assert captured == [2_147_483_647, 1]


def test_recv_json_deframes_decrypts_and_parses(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    # Provide a valid frame_no_crc: [proto][len_lo][len_hi][ciphertext...]
    def _fake_recv_one_frame_no_crc(*, timeout_s: float) -> bytes:
        _ = timeout_s
        return b"\x84\x05\x00" + b"ABCDE"

    monkeypatch.setattr(s, "_recv_one_frame_no_crc", _fake_recv_one_frame_no_crc)

    def _fake_decrypt_schema0_envelope(
        *,
        session_key: bytes,
        protocol_byte: int,
        ciphertext: bytes,
    ) -> _DecryptEnvelope:
        assert s.info is not None
        assert session_key == bytes.fromhex(s.info.session_key_hex)
        assert protocol_byte == 0x84
        assert ciphertext == b"ABCDE"
        return _DecryptEnvelope(payload=b'{"ok":true,"n":2}')

    monkeypatch.setattr(session_mod, "decrypt_schema0_envelope", _fake_decrypt_schema0_envelope)

    obj = s.recv_json(timeout_s=1.0)
    assert obj == {"ok": True, "n": 2}


def test_recv_json_rejects_short_frame(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    def _fake_recv_one_frame_no_crc(*, timeout_s: float) -> bytes:
        _ = timeout_s
        return b"\x80\x00"

    monkeypatch.setattr(s, "_recv_one_frame_no_crc", _fake_recv_one_frame_no_crc)

    # Session now raises SessionProtocolError with a contextual message.
    with pytest.raises(session_mod.SessionProtocolError, match=r"invalid frame \(too short\)"):
        s.recv_json(timeout_s=0.1)


def test_recv_json_rejects_non_object_json(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    def _fake_recv_one_frame_no_crc(*, timeout_s: float) -> bytes:
        _ = timeout_s
        return b"\x81\x03\x00" + b"XYZ"

    monkeypatch.setattr(s, "_recv_one_frame_no_crc", _fake_recv_one_frame_no_crc)

    def _fake_decrypt_schema0_envelope(
        *, session_key: bytes, protocol_byte: int, ciphertext: bytes
    ) -> _DecryptEnvelope:
        _ = session_key, protocol_byte, ciphertext
        return _DecryptEnvelope(payload=b"[1,2,3]")

    monkeypatch.setattr(session_mod, "decrypt_schema0_envelope", _fake_decrypt_schema0_envelope)

    with pytest.raises(session_mod.SessionProtocolError, match=r"Expected a JSON object \(dict\)"):
        s.recv_json(timeout_s=0.2)


def test_recv_json_rejects_invalid_json(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    def _fake_recv_one_frame_no_crc(*, timeout_s: float) -> bytes:
        _ = timeout_s
        return b"\x81\x03\x00" + b"XYZ"

    monkeypatch.setattr(s, "_recv_one_frame_no_crc", _fake_recv_one_frame_no_crc)

    def _fake_decrypt_schema0_envelope(
        *, session_key: bytes, protocol_byte: int, ciphertext: bytes
    ) -> _DecryptEnvelope:
        _ = session_key, protocol_byte, ciphertext
        return _DecryptEnvelope(payload=b'{"unterminated":')

    monkeypatch.setattr(session_mod, "decrypt_schema0_envelope", _fake_decrypt_schema0_envelope)

    with pytest.raises(session_mod.SessionProtocolError, match=r"Received invalid JSON payload"):
        s.recv_json(timeout_s=0.2)


def test_recv_one_frame_no_crc_uses_deframe_state_and_feed(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    # Make _recv_some return two chunks. The first yields no OK frames, second yields an OK frame.
    chunks = [b"chunk1", b"chunk2"]

    def _fake_recv_some(*, max_bytes: int) -> bytes:
        assert max_bytes == s.cfg.recv_max_bytes
        return chunks.pop(0)

    monkeypatch.setattr(s, "_recv_some", _fake_recv_some)

    feed_calls: list[tuple[Any, bytes]] = []

    def _fake_deframe_feed(state: Any, chunk: bytes) -> list[_DeframeResult]:
        # Must be called with canonical state object and the chunk.
        feed_calls.append((state, chunk))
        if chunk == b"chunk1":
            # Simulate CRC-bad frame and/or incomplete parse.
            return [_DeframeResult(ok=False)]
        # Second chunk returns a valid frame.
        return [_DeframeResult(ok=True, frame_no_crc=b"\x80\x01\x00" + b"Z")]

    monkeypatch.setattr(session_mod, "deframe_feed", _fake_deframe_feed)

    recv_one_frame_no_crc = cast(Callable[..., bytes], get_private(s, "_recv_one_frame_no_crc"))
    frame = recv_one_frame_no_crc(timeout_s=0.5)
    assert frame == b"\x80\x01\x00Z"

    # Ensure DeframeState instance was used and deframe_feed called twice.
    assert len(feed_calls) == 2
    deframe_state = cast(DeframeState, get_private(s, "_deframe_state"))
    assert feed_calls[0][0] is deframe_state
    assert feed_calls[0][1] == b"chunk1"
    assert feed_calls[1][0] is deframe_state
    assert feed_calls[1][1] == b"chunk2"


def test_recv_one_frame_no_crc_queues_multiple_frames(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    recv_calls: list[bytes] = []
    chunks = [b"chunk1"]

    def _fake_recv_some(*, max_bytes: int) -> bytes:
        assert max_bytes == s.cfg.recv_max_bytes
        recv_calls.append(b"call")
        return chunks.pop(0)

    monkeypatch.setattr(s, "_recv_some", _fake_recv_some)

    def _fake_deframe_feed(state: Any, chunk: bytes) -> list[_DeframeResult]:
        deframe_state = cast(DeframeState, get_private(s, "_deframe_state"))
        assert state is deframe_state
        assert chunk == b"chunk1"
        return [
            _DeframeResult(ok=True, frame_no_crc=b"\x80\x01\x00A"),
            _DeframeResult(ok=True, frame_no_crc=b"\x81\x01\x00B"),
        ]

    monkeypatch.setattr(session_mod, "deframe_feed", _fake_deframe_feed)

    recv_one_frame_no_crc = cast(Callable[..., bytes], get_private(s, "_recv_one_frame_no_crc"))
    frame1 = recv_one_frame_no_crc(timeout_s=0.5)
    frame2 = recv_one_frame_no_crc(timeout_s=0.5)

    assert frame1 == b"\x80\x01\x00A"
    assert frame2 == b"\x81\x01\x00B"
    assert len(recv_calls) == 1


def test_recv_json_drains_multi_frame_chunk(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    chunks = [b"chunk1"]

    def _fake_recv_some(*, max_bytes: int) -> bytes:
        assert max_bytes == s.cfg.recv_max_bytes
        return chunks.pop(0)

    monkeypatch.setattr(s, "_recv_some", _fake_recv_some)

    def _fake_deframe_feed(state: Any, chunk: bytes) -> list[_DeframeResult]:
        deframe_state = cast(DeframeState, get_private(s, "_deframe_state"))
        assert state is deframe_state
        assert chunk == b"chunk1"
        return [
            _DeframeResult(ok=True, frame_no_crc=b"\x84\x05\x00AAAAA"),
            _DeframeResult(ok=True, frame_no_crc=b"\x84\x05\x00BBBBB"),
        ]

    monkeypatch.setattr(session_mod, "deframe_feed", _fake_deframe_feed)

    payloads = [
        b'{"zone":1}',
        b'{"zone":2}',
    ]

    def _fake_decrypt_schema0_envelope(
        *, session_key: bytes, protocol_byte: int, ciphertext: bytes
    ) -> _DecryptEnvelope:
        _ = session_key, ciphertext
        assert protocol_byte == 0x84
        return _DecryptEnvelope(payload=payloads.pop(0))

    monkeypatch.setattr(session_mod, "decrypt_schema0_envelope", _fake_decrypt_schema0_envelope)

    first = s.recv_json(timeout_s=0.5)
    second = s.recv_json(timeout_s=0.5)

    assert first == {"zone": 1}
    assert second == {"zone": 2}


def test_recv_json_skips_invalid_then_reads_next_frame(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    chunks = [b"chunk1"]

    def _fake_recv_some(*, max_bytes: int) -> bytes:
        assert max_bytes == s.cfg.recv_max_bytes
        return chunks.pop(0)

    monkeypatch.setattr(s, "_recv_some", _fake_recv_some)

    def _fake_deframe_feed(state: Any, chunk: bytes) -> list[_DeframeResult]:
        deframe_state = cast(DeframeState, get_private(s, "_deframe_state"))
        assert state is deframe_state
        assert chunk == b"chunk1"
        return [
            _DeframeResult(ok=True, frame_no_crc=b"\x84\x05\x00AAAAA"),
            _DeframeResult(ok=False),
            _DeframeResult(ok=True, frame_no_crc=b"\x84\x05\x00BBBBB"),
        ]

    monkeypatch.setattr(session_mod, "deframe_feed", _fake_deframe_feed)

    payloads = [
        b'{"zone":3}',
        b'{"zone":4}',
    ]

    def _fake_decrypt_schema0_envelope(
        *, session_key: bytes, protocol_byte: int, ciphertext: bytes
    ) -> _DecryptEnvelope:
        _ = session_key, ciphertext
        assert protocol_byte == 0x84
        return _DecryptEnvelope(payload=payloads.pop(0))

    monkeypatch.setattr(session_mod, "decrypt_schema0_envelope", _fake_decrypt_schema0_envelope)

    first = s.recv_json(timeout_s=0.5)
    second = s.recv_json(timeout_s=0.5)

    assert first == {"zone": 3}
    assert second == {"zone": 4}


def test_recv_json_handles_partial_frame_then_next_chunk(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    chunks = [b"chunk1", b"chunk2"]

    def _fake_recv_some(*, max_bytes: int) -> bytes:
        assert max_bytes == s.cfg.recv_max_bytes
        return chunks.pop(0)

    monkeypatch.setattr(s, "_recv_some", _fake_recv_some)

    def _fake_deframe_feed(state: Any, chunk: bytes) -> list[_DeframeResult]:
        deframe_state = cast(DeframeState, get_private(s, "_deframe_state"))
        assert state is deframe_state
        if chunk == b"chunk1":
            return []
        return [_DeframeResult(ok=True, frame_no_crc=b"\x84\x05\x00HELLO")]

    monkeypatch.setattr(session_mod, "deframe_feed", _fake_deframe_feed)

    def _fake_decrypt_schema0_envelope(
        *, session_key: bytes, protocol_byte: int, ciphertext: bytes
    ) -> _DecryptEnvelope:
        _ = session_key, protocol_byte, ciphertext
        return _DecryptEnvelope(payload=b'{"zone":5}')

    monkeypatch.setattr(session_mod, "decrypt_schema0_envelope", _fake_decrypt_schema0_envelope)

    obj = s.recv_json(timeout_s=0.5)
    assert obj == {"zone": 5}


def test_recv_json_delivers_after_decrypt_error(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    chunks = [b"chunk1"]

    def _fake_recv_some(*, max_bytes: int) -> bytes:
        assert max_bytes == s.cfg.recv_max_bytes
        return chunks.pop(0)

    monkeypatch.setattr(s, "_recv_some", _fake_recv_some)

    def _fake_deframe_feed(state: Any, chunk: bytes) -> list[_DeframeResult]:
        deframe_state = cast(DeframeState, get_private(s, "_deframe_state"))
        assert state is deframe_state
        assert chunk == b"chunk1"
        return [
            _DeframeResult(ok=True, frame_no_crc=b"\x84\x05\x00AAAAA"),
            _DeframeResult(ok=True, frame_no_crc=b"\x84\x05\x00BBBBB"),
        ]

    monkeypatch.setattr(session_mod, "deframe_feed", _fake_deframe_feed)

    def _fake_decrypt_schema0_envelope(
        *, session_key: bytes, protocol_byte: int, ciphertext: bytes
    ) -> _DecryptEnvelope:
        _ = session_key, protocol_byte
        if ciphertext == b"AAAAA":
            raise session_mod.SessionProtocolError("decrypt failed")
        return _DecryptEnvelope(payload=b'{"zone":6}')

    monkeypatch.setattr(session_mod, "decrypt_schema0_envelope", _fake_decrypt_schema0_envelope)

    with pytest.raises(session_mod.SessionProtocolError, match=r"decrypt failed"):
        s.recv_json(timeout_s=0.5)

    obj = s.recv_json(timeout_s=0.5)
    assert obj == {"zone": 6}


def test_pump_once_timeout_returns_none(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    def _fake_recv_json(*, timeout_s: float) -> dict[str, Any]:
        _ = timeout_s
        raise TimeoutError("socket timeout")

    monkeypatch.setattr(s, "recv_json", _fake_recv_json)

    assert s.pump_once(timeout_s=0.01) is None


def test_pump_once_dispatches_on_message(monkeypatch: pytest.MonkeyPatch) -> None:
    s, _ = _make_session_ready(monkeypatch)

    def _fake_recv_json(*, timeout_s: float) -> dict[str, Any]:
        _ = timeout_s
        return {"hello": "world"}

    monkeypatch.setattr(s, "recv_json", _fake_recv_json)

    captured: list[dict[str, Any]] = []

    def _on_message(obj: dict[str, Any]) -> None:
        captured.append(obj)

    s.on_message = _on_message

    out = s.pump_once(timeout_s=0.2)
    assert out == {"hello": "world"}
    assert captured == [{"hello": "world"}]


def test_pump_once_disconnects_and_emits_on_disconnected(monkeypatch: pytest.MonkeyPatch) -> None:
    s, fake_sock = _make_session_ready(monkeypatch)

    class Boom(Exception):
        pass

    def _fake_recv_json(*, timeout_s: float) -> dict[str, Any]:
        _ = timeout_s
        raise Boom("decrypt failed")

    monkeypatch.setattr(s, "recv_json", _fake_recv_json)

    disconnected: dict[str, Any] = {"called": False, "err": None}

    def _on_disconnected(err: Exception | None) -> None:
        disconnected["called"] = True
        disconnected["err"] = err

    s.on_disconnected = _on_disconnected

    with pytest.raises(Boom):
        s.pump_once(timeout_s=0.2)

    # Session should have been closed and callback fired.
    assert s.sock is None
    assert s.info is None
    assert get_private(s, "_deframe_state") is None
    assert fake_sock.closed is True

    assert disconnected["called"] is True
    assert isinstance(disconnected["err"], Boom)



"""
Live E27 test: keypad enumeration via keypads_get_configured + keypads_get_attribs.

Run:
  source ~/elk-e27-env-vars.sh
  pytest -q -m live_e27 test/test_live_e27_keypads_download.py -s
"""

from __future__ import annotations

from typing import cast

import pytest

from elke27_lib.client import Elke27Client
from elke27_lib.const import E27ErrorCode
from test.helpers.error_codes import describe_error, extract_error_code
from test.helpers.payload_validation import assert_payload_shape

_LIVE_TIMEOUT_S = 15.0


@pytest.mark.live_e27
@pytest.mark.asyncio
async def test_live_keypads_download(live_e27_client: Elke27Client) -> None:
    table_info = await live_e27_client.async_execute(
        "keypad_get_table_info", timeout_s=_LIVE_TIMEOUT_S
    )
    if not table_info.ok:
        error_code = extract_error_code(table_info.error)
        if error_code == E27ErrorCode.ELKERR_NOT_READY:
            pytest.skip("Panel reported not-ready for keypad_get_table_info.")
        if error_code == E27ErrorCode.ELKERR_INVALID_TABLE:
            pytest.skip("Keypad table not allocated on this panel.")
        pytest.fail(f"keypad_get_table_info failed: {describe_error(table_info.error)}")
    table_elements = table_info.data.get("table_elements") if table_info.data else None
    if not isinstance(table_elements, int) or table_elements < 1:
        pytest.skip("Keypad table not allocated on this panel.")

    configured = await live_e27_client.async_execute(
        "keypad_get_configured", timeout_s=_LIVE_TIMEOUT_S
    )
    if not configured.ok:
        error_code = extract_error_code(configured.error)
        if error_code == E27ErrorCode.ELKERR_NOT_READY:
            pytest.skip("Panel reported not-ready for keypad_get_configured.")
        if error_code == E27ErrorCode.ELKERR_INVALID_TABLE:
            pytest.fail(
                "keypad_get_configured failed with invalid_table despite allocated keypad table."
            )
        pytest.fail(f"keypad_get_configured failed: {describe_error(configured.error)}")
    assert_payload_shape("keypad_get_configured", configured.data)

    keypad_ids = configured.data.get("keypads") if configured.data else None
    if not isinstance(keypad_ids, list):
        pytest.skip("No configured keypads reported.")

    keypad_ids_list = cast(list[object], keypad_ids)
    keypad_ids = [
        keypad_id for keypad_id in keypad_ids_list if isinstance(keypad_id, int) and keypad_id >= 1
    ]
    if not keypad_ids:
        pytest.skip("No configured keypads reported.")

    for keypad_id in keypad_ids:
        result = await live_e27_client.async_execute(
            "keypad_get_attribs",
            keypad_id=keypad_id,
            timeout_s=_LIVE_TIMEOUT_S,
        )
        if not result.ok and extract_error_code(result.error) == E27ErrorCode.ELKERR_NOT_READY:
            pytest.skip("Panel reported not-ready for keypad_get_attribs.")
        if result.ok:
            assert_payload_shape("keypad_get_attribs", result.data)
            keypad = live_e27_client.state.keypads.get(keypad_id)
            if keypad is not None:
                return

    pytest.skip("Keypad attribs unavailable; auth may be required.")



"""
E27 Session (DDR-0034 / Option A)

Responsibilities:
- Own the TCP socket lifecycle.
- Perform HELLO after connect to obtain session keys.
- Provide a robust framed receive pump using framing.DeframeState + framing.deframe_feed(state, chunk).
- Encrypt+frame outbound schema-0 payloads; deframe+decrypt inbound schema-0 payloads.
- Surface inbound decrypted JSON objects as events (callbacks) or via recv_json().

Non-responsibilities (explicit):
- API_LINK / linking (belongs to provisioning/installer flow).
- Deciding whether/when to AUTHENTICATE (privilege escalation is application-driven).
- Driving application workflows/sequences beyond send/recv primitives.
"""

from __future__ import annotations

import asyncio
import contextlib
import json
import logging
import socket
import threading
import time
from collections import deque
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any, cast

from . import linking
from .errors import E27Error
from .framing import DeframeState, deframe_feed, frame_build
from .hello import perform_hello
from .outbound import OutboundItem, OutboundPriority, OutboundQueue
from .presentation import decrypt_schema0_envelope, encrypt_schema0_envelope

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SessionConfig:
    host: str
    port: int = 2101  # Mitch preference: non-TLS port 2101
    connect_timeout_s: float = 5.0
    io_timeout_s: float = 0.5  # socket read timeout (pump cadence)
    hello_timeout_s: float = 5.0  # overall HELLO timeout
    recv_max_bytes: int = 4096  # per socket recv() call
    protocol_default: int = 0x80  # default protocol byte for schema-0 encrypted frames
    wire_log: bool = False  # enable raw RX/TX hex dump logging
    keepalive_enabled: bool = True
    keepalive_interval_s: float = 30.0
    keepalive_timeout_s: float = 10.0
    keepalive_max_missed: int = 2
    auto_receive: bool = True  # start background receive loop when on_message is set
    auto_receive_thread_fallback: bool = False  # allow dedicated thread when no event loop exists


@dataclass(frozen=True)
class SessionInfo:
    session_id: int
    session_key_hex: str
    session_hmac_hex: str


class SessionState(str, Enum):
    """Internal connection lifecycle states.

    This is intentionally mechanical and policy-free.
    """

    DISCONNECTED = "disconnected"
    CONNECTING = "connecting"
    HELLO = "hello"
    ACTIVE = "active"


class SessionError(RuntimeError):
    """Base exception for Session failures."""


class SessionNotReadyError(SessionError):
    """Raised when an operation requires an ACTIVE session."""


class SessionIOError(SessionError):
    """Raised when the underlying transport fails."""


class SessionProtocolError(SessionError):
    """Raised when framing/crypto/JSON decoding fails."""


class Session:
    """
    Minimal E27 session connection.

    Typical usage:
        s = Session(cfg, client_identity=client_identity, link_key_hex="...")
        s.connect()          # performs HELLO and becomes ready
        s.send_json({...})   # application sends requests (including authenticate if desired)
        obj = s.recv_json()  # or call s.pump_once() to dispatch via callback
    """

    cfg: SessionConfig
    client_identity: linking.E27Identity
    link_key_hex: str
    _tx_envelope_seq: int
    _last_rx_at: float
    _last_tx_at: float
    _last_exchange_at: float
    _rx_count: int
    _recv_lock: threading.Lock

    def __init__(
        self,
        cfg: SessionConfig,
        *,
        client_identity: linking.E27Identity,
        link_key_hex: str,
    ) -> None:
        self.cfg = cfg
        self.client_identity = client_identity
        self.link_key_hex = link_key_hex

        self.sock: socket.socket | None = None
        self._deframe_state: DeframeState | None = None
        self._pending_frames: deque[bytes] = deque()

        self.info: SessionInfo | None = None

        self.state: SessionState = SessionState.DISCONNECTED
        self.last_error: Exception | None = None

        self._tx_envelope_seq = 1
        self._last_rx_envelope_seq: int | None = None
        self._last_rx_json_seq: int | None = None
        self._last_rx_json_keys: tuple[str, ...] | None = None
        self._last_rx_json_domain: str | None = None
        self._last_tx_json_seq: int | None = None
        self._last_tx_json_keys: tuple[str, ...] | None = None
        self._last_tx_json_domain: str | None = None
        now = time.monotonic()
        self._last_rx_at = now
        self._last_tx_at = now
        self._last_exchange_at = now
        self._rx_count = 0
        self._recv_thread: threading.Thread | None = None
        self._recv_stop: threading.Event | None = None
        self._recv_lock = threading.Lock()
        self._recv_task: asyncio.Task[None] | None = None
        self._recv_loop_ref: asyncio.AbstractEventLoop | None = None
        self._outbound: OutboundQueue | None = None

        # Event hooks (optional)
        self.on_connected: Callable[[SessionInfo], None] | None = None
        self.on_message: Callable[[dict[str, Any]], None] | None = None
        self.on_disconnected: Callable[[Exception | None], None] | None = None
        self.on_idle: Callable[[], None] | None = None

    # --------------------------
    # Connection lifecycle
    # --------------------------

    def connect(self) -> SessionInfo:
        """
        Connect TCP and perform HELLO to obtain session keys.
        """
        if self.state is not SessionState.DISCONNECTED:
            # Mechanical safety: connect() is intended to establish a new session.
            self.close()

        self.last_error = None
        self.state = SessionState.CONNECTING

        logger.info("E27 Session connecting to %s:%s", self.cfg.host, self.cfg.port)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.settimeout(self.cfg.connect_timeout_s)
            s.connect((self.cfg.host, self.cfg.port))
        except OSError as e:
            self.last_error = e
            self.state = SessionState.DISCONNECTED
            with contextlib.suppress(OSError):
                s.close()
            raise SessionIOError(
                f"Failed to connect to {self.cfg.host}:{self.cfg.port}: {e}"
            ) from e

        # After connect, switch to pump cadence timeout.
        s.settimeout(self.cfg.io_timeout_s)
        self.sock = s
        self._deframe_state = DeframeState()

        self.state = SessionState.HELLO
        try:
            keys = perform_hello(
                sock=s,
                client_identity=self.client_identity,
                linkkey_hex=self.link_key_hex,
                timeout_s=self.cfg.hello_timeout_s,
            )
        except E27Error as e:
            self.last_error = e
            # HELLO failure is a session setup failure; close and surface clearly.
            self._handle_disconnect(e)
            raise SessionProtocolError(
                f"HELLO failed for {self.cfg.host}:{self.cfg.port}: {e}"
            ) from e

        self.info = SessionInfo(
            session_id=keys.session_id,
            session_key_hex=keys.session_key_hex,
            session_hmac_hex=keys.hmac_key_hex,
        )
        self.state = SessionState.ACTIVE
        self._tx_envelope_seq = 1
        self._last_rx_at = time.monotonic()
        self._last_tx_at = self._last_rx_at
        self._last_exchange_at = self._last_rx_at

        logger.info("E27 HELLO complete; session_id=%s", self.info.session_id)

        if self.on_connected:
            self.on_connected(self.info)

        if self.cfg.auto_receive and self.on_message is not None:
            self._start_receiver()

        return self.info

    def close(self) -> None:
        """
        Close the socket. Safe to call multiple times.
        """
        self._stop_receiver()
        if self._outbound is not None:
            self._outbound.stop(fail_exc=SessionIOError("Session closed."))
            self._outbound = None
        if self.sock is not None:
            with contextlib.suppress(OSError):
                self.sock.close()
        self.sock = None
        self._deframe_state = None
        self._pending_frames = deque()
        self.info = None
        self.state = SessionState.DISCONNECTED

    def handle_disconnect(self, err: Exception | None) -> None:
        self._handle_disconnect(err)

    # --------------------------
    # Transport helpers
    # --------------------------

    def _require_ready(self) -> None:
        if (
            self.state is not SessionState.ACTIVE
            or self.sock is None
            or self.info is None
            or self._deframe_state is None
        ):
            raise SessionNotReadyError(
                "Session is not ACTIVE/ready (call connect() successfully first)."
            )

    def _recv_some(self, *, max_bytes: int) -> bytes:
        """
        Read from socket; may raise TimeoutError or ConnectionError.
        Kept as a method so tests can monkeypatch it.
        """
        self._require_ready()
        assert self.sock is not None

        try:
            data = self.sock.recv(max_bytes)
        except TimeoutError as e:
            raise TimeoutError("Timed out waiting for data from the panel.") from e
        except OSError as e:
            raise SessionIOError(
                f"Socket read failed from {self.cfg.host}:{self.cfg.port}: {e}"
            ) from e

        if not data:
            raise SessionIOError(
                f"Connection closed by the panel ({self.cfg.host}:{self.cfg.port})."
            )

        return data

    def _send_all(self, data: bytes) -> None:
        self._require_ready()
        assert self.sock is not None
        try:
            self.sock.sendall(data)
            now = time.monotonic()
            self._last_tx_at = now
            self._last_exchange_at = now
        except OSError as e:
            raise SessionIOError(
                f"Socket write failed to {self.cfg.host}:{self.cfg.port}: {e}"
            ) from e

    # --------------------------
    # Framed receive pump
    # --------------------------

    def _recv_one_frame_no_crc(self, *, timeout_s: float) -> bytes:
        """
        Return the first valid frame_no_crc from the stream.

        frame_no_crc layout (per framing.deframe_feed):
            [protocol_byte][len_lo][len_hi][ciphertext...]
        """
        self._require_ready()
        assert self._deframe_state is not None
        if self._pending_frames:
            return self._pending_frames.popleft()

        deadline = time.monotonic() + timeout_s
        while True:
            remaining = deadline - time.monotonic()
            if remaining <= 0:
                raise TimeoutError("Timed out waiting for a framed message from the panel.")

            try:
                chunk = self._recv_some(max_bytes=self.cfg.recv_max_bytes)
            except TimeoutError:
                # keep pumping until overall deadline; allow idle hooks for retries
                if self.on_idle:
                    with contextlib.suppress(Exception):
                        self.on_idle()
                continue

            if self.cfg.wire_log and logger.isEnabledFor(logging.DEBUG):
                logger.debug("RX raw chunk (%d bytes): %s", len(chunk), chunk.hex())

            results = deframe_feed(self._deframe_state, chunk)
            for r in results:
                if getattr(r, "ok", False):
                    if r.frame_no_crc is None:
                        continue
                    self._pending_frames.append(r.frame_no_crc)
                # CRC-bad or malformed frames: ignore and keep scanning.
                # If the framing layer provides details, emit at debug level.
                err = getattr(r, "error", None)
                if err:
                    logger.warning("Dropping invalid frame while resyncing: %s", err)
            if self._pending_frames:
                frame = self._pending_frames.popleft()
                if self.cfg.wire_log and logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "RX frame_no_crc (%d bytes): %s",
                        len(frame),
                        frame.hex(),
                    )
                return frame

    # --------------------------
    # Public send/recv API
    # --------------------------

    def send_json(
        self,
        obj: dict[str, Any],
        *,
        priority: OutboundPriority = OutboundPriority.NORMAL,
        on_sent: Callable[[float], None] | None = None,
        on_fail: Callable[[BaseException], None] | None = None,
    ) -> None:
        """
        Encrypt schema-0 payload (JSON UTF-8 bytes) and send as a framed message.
        """
        self._note_tx_json(obj)
        framed = self._encode_json(obj)
        if self._outbound is None:
            self._send_all(framed)
            if on_sent is not None:
                on_sent(time.monotonic())
            return

        item = OutboundItem(
            payload=framed,
            seq=obj.get("seq") if isinstance(obj.get("seq"), int) else None,
            kind="request",
            priority=priority,
            enqueued_at=time.monotonic(),
            on_sent=on_sent,
            on_fail=on_fail,
        )
        self._outbound.enqueue(item)

    def enable_outbound_queue(
        self,
        *,
        loop: asyncio.AbstractEventLoop,
        min_interval_s: float,
        max_burst: int,
    ) -> None:
        self._outbound = OutboundQueue(
            loop=loop,
            send_fn=self._send_all,
            min_interval_s=min_interval_s,
            max_burst=max_burst,
            logger=logger,
        )
        self._outbound.start()

    def _encode_json(self, obj: dict[str, Any]) -> bytes:
        self._require_ready()
        assert self.info is not None

        payload = json.dumps(obj, separators=(",", ":"), ensure_ascii=False).encode("utf-8")

        envelope_seq = self._tx_envelope_seq
        self._tx_envelope_seq = self._next_envelope_seq(envelope_seq)
        proto, ciphertext = encrypt_schema0_envelope(
            payload=payload,
            session_key=bytes.fromhex(self.info.session_key_hex),
            src=1,
            dest=0,
            head=0,
            envelope_seq=envelope_seq,
        )

        framed = frame_build(protocol_byte=proto, data_frame=ciphertext)
        if self.cfg.wire_log and logger.isEnabledFor(logging.DEBUG):
            logger.debug("TX framed (%d bytes): %s", len(framed), framed.hex())
        return framed

    def recv_json(self, *, timeout_s: float = 5.0) -> dict[str, Any]:
        """
        Receive one framed message, decrypt schema-0, parse JSON, return dict.
        """
        with self._recv_lock:
            self._require_ready()
            assert self.info is not None
            idle_check_at = time.monotonic() + timeout_s
            while True:
                if self.on_idle and time.monotonic() >= idle_check_at:
                    with contextlib.suppress(Exception):
                        self.on_idle()
                    idle_check_at = time.monotonic() + timeout_s
                frame_no_crc = self._recv_one_frame_no_crc(timeout_s=timeout_s)
                if len(frame_no_crc) < 3:
                    logger.warning(
                        "Dropping short frame (len=%d) from %s:%s",
                        len(frame_no_crc),
                        self.cfg.host,
                        self.cfg.port,
                    )
                    raise SessionProtocolError(
                        f"Received an invalid frame (too short) from {self.cfg.host}:{self.cfg.port}."
                    )

                protocol_byte = frame_no_crc[0]
                frame_len = frame_no_crc[1] | (frame_no_crc[2] << 8)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "RX frame header: protocol=0x%02x length=%d total=%d",
                        protocol_byte,
                        frame_len,
                        len(frame_no_crc),
                    )
                ciphertext = frame_no_crc[3:]  # skip protocol + 2-byte length

                try:
                    env = decrypt_schema0_envelope(
                        protocol_byte=protocol_byte,
                        ciphertext=ciphertext,
                        session_key=bytes.fromhex(self.info.session_key_hex),
                    )
                except Exception as e:
                    logger.warning(
                        "Dropping frame after decrypt failure: protocol=0x%02x length=%d ciphertext_len=%d error=%s",
                        protocol_byte,
                        frame_len,
                        len(ciphertext),
                        e,
                    )
                    raise SessionProtocolError(
                        f"Failed to decrypt schema-0 envelope from {self.cfg.host}:{self.cfg.port}: {e}"
                    ) from e
                seq_val = getattr(env, "seq", None)
                if isinstance(seq_val, int):
                    self._last_rx_envelope_seq = seq_val

                try:
                    obj = json.loads(env.payload.decode("utf-8"))
                except Exception as e:
                    logger.warning(
                        "Dropping frame after JSON decode failure: protocol=0x%02x length=%d error=%s",
                        protocol_byte,
                        frame_len,
                        e,
                    )
                    raise SessionProtocolError(
                        f"Received invalid JSON payload from {self.cfg.host}:{self.cfg.port}: {e}"
                    ) from e

                if not isinstance(obj, dict):
                    logger.warning(
                        "Dropping non-object JSON payload: protocol=0x%02x length=%d type=%s",
                        protocol_byte,
                        frame_len,
                        type(obj).__name__,
                    )
                    raise SessionProtocolError(
                        "Expected a JSON object (dict) but received "
                        f"{type(obj).__name__} from {self.cfg.host}:{self.cfg.port}."
                    )

                obj = cast(dict[str, Any], obj)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "RX json: session_id=%s seq=%s keys=%s",
                        obj.get("session_id"),
                        obj.get("seq"),
                        tuple(obj.keys()),
                    )
                    logger.debug(
                        "RX json decoded: domain=%s",
                        self._extract_domain_key(obj),
                    )
                self._note_rx_json(obj)
                self._rx_count += 1
                self._last_rx_at = time.monotonic()
                self._last_exchange_at = self._last_rx_at
                return obj

    def pump_once(self, *, timeout_s: float = 0.5) -> dict[str, Any] | None:
        """
        One pump iteration: receive and dispatch exactly one message if available.

        Returns:
            The decoded JSON dict if one was received, else None on timeout.
        """
        try:
            obj = self.recv_json(timeout_s=timeout_s)
        except TimeoutError:
            if self.on_idle:
                self.on_idle()
            return None
        except SessionNotReadyError:
            # Caller attempted to pump without a connected session.
            raise
        except (SessionIOError, SessionProtocolError) as e:
            # A transport/protocol failure means the session is no longer healthy.
            logger.warning(
                "Session pump failed (%s) in state=%s for %s:%s: %s",
                type(e).__name__,
                self.state.value,
                self.cfg.host,
                self.cfg.port,
                e,
            )
            self._handle_disconnect(e)
            raise
        except Exception as e:
            # Unexpected error: still treat as disconnect-worthy at the Session layer.
            logger.warning(
                "Unexpected session pump error (%s) in state=%s for %s:%s: %s",
                type(e).__name__,
                self.state.value,
                self.cfg.host,
                self.cfg.port,
                e,
            )
            self._handle_disconnect(e)
            raise

        if self.on_message:
            self.on_message(obj)

        return obj

    def _start_receiver(self) -> None:
        if self._recv_thread is not None or self._recv_task is not None:
            return
        self._recv_stop = threading.Event()
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop is not None:
            # Prefer asyncio.to_thread when a loop is running (HA async contexts).
            self._recv_loop_ref = loop
            self._recv_task = loop.create_task(asyncio.to_thread(self._recv_loop, self._recv_stop))
            return

        if not self.cfg.auto_receive_thread_fallback:
            return

        self._recv_thread = threading.Thread(
            target=self._recv_loop,
            args=(self._recv_stop,),
            name="e27-recv",
            daemon=True,
        )
        self._recv_thread.start()

    def start_auto_receive(self) -> None:
        """Start auto-receive if enabled and a message handler is configured."""
        if (
            not self.cfg.auto_receive
            or self.on_message is None
            or self.state is not SessionState.ACTIVE
        ):
            return
        self._start_receiver()

    def _stop_receiver(self) -> None:
        if self._recv_stop is not None:
            self._recv_stop.set()
        if self._recv_thread is not None and self._recv_thread is not threading.current_thread():
            self._recv_thread.join(timeout=1.0)
        if self._recv_task is not None:
            # Let the to_thread worker exit via stop_event; no hard cancel needed.
            self._recv_task = None
            self._recv_loop_ref = None
        self._recv_thread = None
        self._recv_stop = None

    def _recv_loop(self, stop_event: threading.Event) -> None:
        while not stop_event.is_set():
            if self.state is not SessionState.ACTIVE:
                stop_event.wait(0.1)
                continue
            try:
                obj = self.recv_json(timeout_s=self.cfg.io_timeout_s)
            except TimeoutError:
                if self.on_idle:
                    with contextlib.suppress(Exception):
                        self.on_idle()
                continue
            except SessionNotReadyError:
                break
            except (SessionIOError, SessionProtocolError) as e:
                self._handle_disconnect(e)
                break
            except Exception as e:
                self._handle_disconnect(e)
                logger.warning("Session receive loop error: %s", e, exc_info=True)
                break

            if self.on_message:
                self.on_message(obj)

    def _handle_disconnect(self, err: Exception | None) -> None:
        now = time.monotonic()
        rx_age = now - self._last_rx_at
        tx_age = now - self._last_tx_at
        exchange_age = now - self._last_exchange_at
        err_name = type(err).__name__ if err is not None else "None"
        logger.error(
            "Session disconnect: err=%s state=%s host=%s port=%s rx_age=%.3fs tx_age=%.3fs "
            "exchange_age=%.3fs rx_count=%s last_rx_seq=%s last_rx_domain=%s last_tx_seq=%s "
            "last_tx_domain=%s",
            err_name,
            self.state.value,
            self.cfg.host,
            self.cfg.port,
            rx_age,
            tx_age,
            exchange_age,
            self._rx_count,
            self._last_rx_json_seq,
            self._last_rx_json_domain,
            self._last_tx_json_seq,
            self._last_tx_json_domain,
        )
        self.last_error = err
        try:
            self.close()
        finally:
            if self.on_disconnected:
                self.on_disconnected(err)

    @staticmethod
    def _extract_domain_key(obj: dict[str, Any]) -> str | None:
        for key in obj.keys():
            if key not in ("seq", "session_id"):
                return key
        return None

    def _note_rx_json(self, obj: dict[str, Any]) -> None:
        seq_val = obj.get("seq")
        if isinstance(seq_val, int):
            self._last_rx_json_seq = seq_val
        self._last_rx_json_keys = tuple(obj.keys())
        self._last_rx_json_domain = self._extract_domain_key(obj)

    def _note_tx_json(self, obj: dict[str, Any]) -> None:
        seq_val = obj.get("seq")
        if isinstance(seq_val, int):
            self._last_tx_json_seq = seq_val
        self._last_tx_json_keys = tuple(obj.keys())
        self._last_tx_json_domain = self._extract_domain_key(obj)

    def _next_envelope_seq(self, current: int) -> int:
        return self._wrap_seq(current + 1, wrap_to=1)

    @staticmethod
    def _wrap_seq(value: int, *, wrap_to: int) -> int:
        max_seq = 2_147_483_647
        if value > max_seq:
            return wrap_to
        if value <= 0:
            return wrap_to
        return value

    def reconnect(self) -> SessionInfo:
        """Mechanical reconnect helper (no backoff/policy).

        This is intentionally a convenience wrapper around close() + connect().
        Any retry/backoff strategy belongs above the Session layer.
        """
        self.close()
        return self.connect()



from collections.abc import Mapping
from typing import cast

from elke27_lib.const import E27ErrorCode
from elke27_lib.dispatcher import (
    ERR_DOMAIN_EMPTY,
    ERR_DOMAIN_MULTI,
    ERR_INVALID_SEQ,
    ERR_ROOT_EMPTY,
    ERR_ROOT_MULTI,
    ERR_UNEXPECTED_VALUE_TYPE,
    ERROR_ALL,
    ERROR_DOMAIN,
    ERROR_ROOT,
    DispatchContext,
    Dispatcher,
    MessageKind,
    PagedBlock,
    PagedTransferKey,
    PendingRequest,
)

# ---------------------------------------------------------------------------
# helpers
# ---------------------------------------------------------------------------


class Recorder:
    def __init__(self) -> None:
        self.calls: list[tuple[Mapping[str, object], DispatchContext]] = []

    def handler(self, msg: Mapping[str, object], ctx: DispatchContext) -> bool:
        self.calls.append((msg, ctx))
        return True


def _zone_empty_msg(seq: int) -> dict[str, object]:
    zone_payload: dict[str, object] = {}
    return {"zone": zone_payload, "seq": seq}


def _zone_status_msg(seq: int | str | None = None) -> dict[str, object]:
    status_payload: dict[str, object] = {}
    zone_payload: dict[str, object] = {"status": status_payload}
    msg: dict[str, object] = {"zone": zone_payload}
    if seq is not None:
        msg["seq"] = seq
    return msg


# ---------------------------------------------------------------------------
# routing tests
# ---------------------------------------------------------------------------


def test_simple_domain_and_command_route():
    d = Dispatcher()
    r = Recorder()
    d.register(("zone", "status"), r.handler)

    msg: dict[str, object] = {"zone": {"status": {"id": 1}}, "seq": 1}
    result = d.dispatch(msg)

    assert result.route == ("zone", "status")
    assert result.kind == MessageKind.DIRECTED
    assert result.classification == "UNSOLICITED"
    assert result.handled is True
    assert len(result.errors) == 0
    assert len(r.calls) == 1


def test_domain_empty_dict_routes_to_empty_and_emits_error():
    d = Dispatcher()
    r = Recorder()
    e = Recorder()

    d.register(("zone", "__root__"), r.handler)
    d.register((ERROR_DOMAIN, ERR_DOMAIN_EMPTY), e.handler)

    msg = _zone_empty_msg(1)
    result = d.dispatch(msg)

    assert result.route == ("zone", "__empty__")
    assert any(err.code == ERR_DOMAIN_EMPTY for err in result.errors)
    assert len(r.calls) == 1
    assert len(e.calls) == 1
    err_msg = e.calls[0][0]
    error_domain = cast(Mapping[str, Mapping[str, object]], err_msg["__error__"])
    payload = error_domain[ERR_DOMAIN_EMPTY].get("payload")
    assert isinstance(payload, str)
    assert '"zone":{}' in payload


def test_domain_multi_key_routes_to_root():
    d = Dispatcher()
    r = Recorder()

    d.register(("zone", "__root__"), r.handler)

    msg: dict[str, object] = {"zone": {"a": 1, "b": 2}, "seq": 1}
    result = d.dispatch(msg)

    assert result.route == ("zone", "__root__")
    assert any(err.code == ERR_DOMAIN_MULTI for err in result.errors)
    assert len(r.calls) == 1


def test_domain_root_error_routes_to_error_without_dispatch_error():
    d = Dispatcher()

    msg: dict[str, object] = {
        "seq": 1,
        "network": {"error_code": E27ErrorCode.ELKERR_NOAUTH, "error_message": "no authorization"},
    }
    result = d.dispatch(msg)

    assert result.route == ("network", "error")
    assert not any(err.code == ERR_DOMAIN_MULTI for err in result.errors)


def test_domain_unexpected_value_type():
    d = Dispatcher()
    r = Recorder()

    d.register(("zone", "__root__"), r.handler)

    msg: dict[str, object] = {"zone": True, "seq": 1}
    result = d.dispatch(msg)

    assert result.route == ("zone", "__value__")
    assert any(err.code == ERR_UNEXPECTED_VALUE_TYPE for err in result.errors)
    assert len(r.calls) == 1


def test_root_empty():
    d = Dispatcher()
    e = Recorder()

    d.register((ERROR_DOMAIN, ERR_ROOT_EMPTY), e.handler)

    msg: dict[str, object] = {"seq": 1}
    result = d.dispatch(msg)

    assert result.route == ("__root__", "__empty__")
    assert any(err.code == ERR_ROOT_EMPTY for err in result.errors)
    assert len(e.calls) == 1
    err_msg = e.calls[0][0]
    error_domain = cast(Mapping[str, Mapping[str, object]], err_msg["__error__"])
    payload = error_domain[ERR_ROOT_EMPTY].get("payload")
    assert isinstance(payload, str)
    assert '"seq":1' in payload


def test_root_multi_domain():
    d = Dispatcher()
    e = Recorder()

    d.register((ERROR_DOMAIN, ERR_ROOT_MULTI), e.handler)

    area_payload: dict[str, object] = {}
    zone_payload: dict[str, object] = {}
    msg: dict[str, object] = {"zone": zone_payload, "area": area_payload, "seq": 1}
    result = d.dispatch(msg)

    assert result.route == ("__root__", "__multi__")
    assert any(err.code == ERR_ROOT_MULTI for err in result.errors)
    assert len(e.calls) == 1


def test_root_error_envelope_routes_to_panel_error():
    d = Dispatcher()
    r = Recorder()

    d.register((ERROR_DOMAIN, ERROR_ROOT), r.handler)

    msg: dict[str, object] = {"seq": 0, "error_code": 11008, "error_message": "no authorization"}
    result = d.dispatch(msg)

    assert result.route == (ERROR_DOMAIN, ERROR_ROOT)
    assert result.kind == MessageKind.BROADCAST
    assert result.classification == "BROADCAST"
    assert result.handled is True
    assert not any(err.code == ERR_ROOT_MULTI for err in result.errors)
    assert len(r.calls) == 1


# ---------------------------------------------------------------------------
# seq + classification tests
# ---------------------------------------------------------------------------


def test_missing_seq_is_unknown_without_error():
    d = Dispatcher()
    r = Recorder()
    d.register(("zone", "status"), r.handler)

    msg = _zone_status_msg()
    result = d.dispatch(msg)

    assert result.kind == MessageKind.UNKNOWN
    assert result.classification == "UNKNOWN"
    assert len(result.errors) == 0


def test_broadcast_seq_zero():
    d = Dispatcher()
    r = Recorder()
    d.register(("zone", "status"), r.handler)

    msg = _zone_status_msg(0)
    result = d.dispatch(msg)

    assert result.kind == MessageKind.BROADCAST
    assert result.classification == "BROADCAST"
    assert len(r.calls) == 1


def test_invalid_seq_type_emits_error():
    d = Dispatcher()
    e = Recorder()

    d.register((ERROR_DOMAIN, ERR_INVALID_SEQ), e.handler)

    msg = _zone_status_msg("abc")
    result = d.dispatch(msg)

    assert any(err.code == ERR_INVALID_SEQ for err in result.errors)
    assert len(e.calls) == 1


def test_negative_seq_emits_error_and_unknown():
    d = Dispatcher()
    e = Recorder()

    d.register((ERROR_DOMAIN, ERR_INVALID_SEQ), e.handler)

    msg = _zone_status_msg(-1)
    result = d.dispatch(msg)

    assert result.kind == MessageKind.UNKNOWN
    assert any(err.code == ERR_INVALID_SEQ for err in result.errors)
    assert len(e.calls) == 1


# ---------------------------------------------------------------------------
# pending correlation tests
# ---------------------------------------------------------------------------


def test_pending_request_is_matched_and_popped():
    d = Dispatcher()
    r = Recorder()

    d.register(("zone", "status"), r.handler)

    pending = PendingRequest(seq=42)
    d.add_pending(pending)

    msg = _zone_status_msg(42)
    result = d.dispatch(msg)

    assert result.classification == "RESPONSE"
    assert result.response_match is pending
    assert d.match_pending(42, pop=False) is None


def test_directed_without_pending_is_unsolicited():
    d = Dispatcher()

    msg = _zone_status_msg(99)
    result = d.dispatch(msg)

    assert result.classification == "UNSOLICITED"


# ---------------------------------------------------------------------------
# error envelope tests
# ---------------------------------------------------------------------------


def test_error_envelope_shape_and_raw_route():
    d = Dispatcher()
    e = Recorder()

    d.register((ERROR_DOMAIN, ERR_DOMAIN_EMPTY), e.handler)
    d.register((ERROR_DOMAIN, ERROR_ALL), e.handler)

    msg = _zone_empty_msg(5)
    d.dispatch(msg)

    assert len(e.calls) == 2

    err_msg, ctx = e.calls[0]

    assert ERROR_DOMAIN in err_msg
    error_domain = cast(Mapping[str, Mapping[str, object]], err_msg[ERROR_DOMAIN])
    code = next(iter(error_domain))
    payload = error_domain[code]

    assert payload["message"]
    assert payload["domain"] == "zone"
    assert payload["severity"]
    assert ctx.raw_route == ("zone", "__empty__")
    assert ctx.route == (ERROR_DOMAIN, ERR_DOMAIN_EMPTY)


# ---------------------------------------------------------------------------
# handler isolation
# ---------------------------------------------------------------------------


def test_handler_exception_does_not_break_dispatch():
    d = Dispatcher()
    good = Recorder()

    def bad_handler(_msg: Mapping[str, object], _ctx: DispatchContext) -> bool:
        raise RuntimeError("boom")

    d.register(("zone", "status"), bad_handler)
    d.register(("zone", "status"), good.handler)

    msg = _zone_status_msg(1)
    result = d.dispatch(msg)

    assert result.handled is True
    assert len(good.calls) == 1


def test_paged_reassembly_out_of_order() -> None:
    d = Dispatcher()
    r = Recorder()
    route = ("zone", "get_configured")

    def _merge(blocks: list[PagedBlock], block_count: int) -> dict[str, object]:
        merged: list[int] = []
        for block in blocks:
            zones = block.payload.get("zones", [])
            if isinstance(zones, list):
                merged.extend(cast(list[int], zones))
        return {"zones": merged, "block_count": block_count}

    d.register_paged(route, merge_fn=_merge)
    d.register(route, r.handler)

    key = PagedTransferKey(session_id=1, transfer_id=99, route=route)
    d.add_pending(PendingRequest(seq=10, opaque=key))
    d.add_pending(PendingRequest(seq=11, opaque=key))

    msg2 = {
        "seq": 11,
        "session_id": 1,
        "zone": {"get_configured": {"block_id": 2, "block_count": 2, "zones": [3]}},
    }
    d.dispatch(msg2)
    assert len(r.calls) == 0

    msg1 = {
        "seq": 10,
        "session_id": 1,
        "zone": {"get_configured": {"block_id": 1, "block_count": 2, "zones": [1, 2]}},
    }
    d.dispatch(msg1)
    assert len(r.calls) == 1
    assembled = cast(Mapping[str, Mapping[str, Mapping[str, object]]], r.calls[0][0])
    zones = cast(
        Mapping[str, object],
        cast(Mapping[str, object], assembled["zone"])["get_configured"],
    )["zones"]
    assert zones == [1, 2, 3]



"""
elke27_lib/handlers/keypad.py

Read-only handlers for the "keypad" domain.
"""

from __future__ import annotations

import logging
from collections.abc import Callable, Mapping
from typing import Any, cast

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    ApiError,
    AuthorizationRequiredEvent,
    CsmSnapshotUpdated,
    Event,
    KeypadConfiguredInventoryReady,
    TableCsmChanged,
)
from elke27_lib.states import KeypadState, PanelState, update_csm_snapshot

EmitFn = Callable[[Event, DispatchContext], None]
NowFn = Callable[[], float]

LOG = logging.getLogger(__name__)


def _as_mapping(obj: object) -> Mapping[str, Any] | None:
    if isinstance(obj, Mapping):
        return cast(Mapping[str, Any], obj)
    return None


def _coerce_int(value: object) -> int | None:
    return value if isinstance(value, int) else None


def make_keypad_get_configured_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("keypad","get_configured").
    """

    def handler_keypad_get_configured(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        keypad_obj = _as_mapping(msg.get("keypad"))
        if keypad_obj is None:
            return False

        payload = _as_mapping(keypad_obj.get("get_configured"))
        if payload is None:
            return False

        error_code = payload.get("error_code", keypad_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="keypad",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="keypad",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        keypads = _extract_configured_ids(payload)
        if keypads:
            state.inventory.configured_keypads.update(keypads)
            for keypad_id in keypads:
                state.get_or_create_keypad(keypad_id)

        block_id = payload.get("block_id")
        block_count = payload.get("block_count")
        if (
            isinstance(block_id, int)
            and isinstance(block_count, int)
            and block_count >= 1
            and block_id >= block_count
        ):
            state.inventory.configured_keypads_complete = True
            emit(
                KeypadConfiguredInventoryReady(
                    kind=KeypadConfiguredInventoryReady.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                ),
                ctx,
            )

        state.panel.last_message_at = now()
        return True

    return handler_keypad_get_configured


def make_keypad_get_attribs_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("keypad","get_attribs").
    """

    def handler_keypad_get_attribs(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        keypad_obj = _as_mapping(msg.get("keypad"))
        if keypad_obj is None:
            return False

        payload = _as_mapping(keypad_obj.get("get_attribs"))
        if payload is None:
            return False

        error_code = payload.get("error_code", keypad_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="keypad",
                        entity_id=_coerce_int(payload.get("keypad_id")),
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="keypad",
                    entity_id=_coerce_int(payload.get("keypad_id")),
                    message=None,
                ),
                ctx,
            )
            return True

        keypad_id = payload.get("keypad_id")
        if not isinstance(keypad_id, int) or keypad_id < 1:
            return False

        keypad = state.get_or_create_keypad(keypad_id)
        changed: set[str] = set()
        _apply_keypad_attribs(keypad, payload, changed)
        keypad.last_update_at = now()
        state.panel.last_message_at = keypad.last_update_at
        return True

    return handler_keypad_get_attribs


def make_keypad_get_table_info_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("keypad","get_table_info").
    """

    def handler_keypad_get_table_info(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        keypad_obj = _as_mapping(msg.get("keypad"))
        if keypad_obj is None:
            return False

        payload = _as_mapping(keypad_obj.get("get_table_info"))
        if payload is None:
            payload = _as_mapping(keypad_obj.get("table_info"))
        if payload is None:
            return False

        error_code = payload.get("error_code")
        if isinstance(error_code, int) and error_code != 0:
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="keypad",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        state.table_info_by_domain["keypad"] = dict(payload)
        state.panel.last_message_at = now()
        table_csm = _extract_table_csm(payload, domain="keypad")
        if table_csm is not None:
            old = state.table_csm_by_domain.get("keypad")
            if old != table_csm:
                state.table_csm_by_domain["keypad"] = table_csm
                emit(
                    TableCsmChanged(
                        kind=TableCsmChanged.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        csm_domain="keypad",
                        old=old,
                        new=table_csm,
                    ),
                    ctx,
                )

        snapshot = update_csm_snapshot(state)
        if snapshot is not None:
            emit(
                CsmSnapshotUpdated(
                    kind=CsmSnapshotUpdated.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    snapshot=snapshot,
                ),
                ctx,
            )
        return True

    return handler_keypad_get_table_info


def _extract_configured_ids(payload: Mapping[str, Any]) -> set[int]:
    for key in ("keypads", "keypad_ids", "configured_keypads", "configured_keypad_ids"):
        value = payload.get(key)
        if isinstance(value, list):
            return {
                item for item in cast(list[object], value) if isinstance(item, int) and item >= 1
            }
    return set()


def _normalize_name(value: Any) -> str | None:
    if value is None:
        return None
    text = str(value).strip()
    return text if text else None


def _apply_keypad_attribs(
    keypad: KeypadState, payload: Mapping[str, Any], changed: set[str]
) -> None:
    if "name" in payload:
        name = _normalize_name(payload.get("name"))
        if keypad.name != name:
            keypad.name = name
            changed.add("name")
    if "area" in payload:
        area = payload.get("area")
        if isinstance(area, int) and keypad.area != area:
            keypad.area = area
            changed.add("area")
    if "zone_id" in payload:
        zone_id = payload.get("zone_id")
        if isinstance(zone_id, int) and keypad.zone_id != zone_id:
            keypad.zone_id = zone_id
            changed.add("zone_id")
    if "source_id" in payload:
        source_id = payload.get("source_id")
        if isinstance(source_id, int) and keypad.source_id != source_id:
            keypad.source_id = source_id
            changed.add("source_id")
    if "device_id" in payload:
        device_id = payload.get("device_id")
        if isinstance(device_id, str) and keypad.device_id != device_id:
            keypad.device_id = device_id
            changed.add("device_id")
    if "flags" in payload:
        flags = payload.get("flags")
        if isinstance(flags, list) and keypad.flags != flags:
            keypad.flags = flags
            changed.add("flags")

    for key, value in payload.items():
        if key in {
            "keypad_id",
            "error_code",
            "name",
            "area",
            "zone_id",
            "source_id",
            "device_id",
            "flags",
        }:
            continue
        if keypad.fields.get(key) != value:
            keypad.fields[key] = value
            changed.add(key)


def _extract_table_csm(payload: Mapping[str, Any], *, domain: str) -> int | None:
    if "table_csm" not in payload:
        return None
    value = payload.get("table_csm")
    if isinstance(value, bool):
        LOG.warning("%s.get_table_info table_csm has invalid bool value.", domain)
        return None
    if isinstance(value, int):
        return value
    if isinstance(value, float) and value.is_integer():
        return int(value)
    if isinstance(value, str):
        text = value.strip()
        if text.isdigit():
            return int(text)
    LOG.warning("%s.get_table_info table_csm has non-int value %r.", domain, value)
    return None



from __future__ import annotations

from elke27_lib.const import E27ErrorCode
from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import Event
from elke27_lib.handlers.system import make_system_get_trouble_handler
from elke27_lib.states import PanelState
from test.helpers.dispatch import make_ctx


class _EmitSpy:
    def __init__(self) -> None:
        self.events: list[Event] = []

    def __call__(self, evt: Event, _ctx: DispatchContext) -> None:
        self.events.append(evt)


_Ctx = make_ctx


def test_system_get_trouble_sets_status_snapshot() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_system_get_trouble_handler(state, emit, now=lambda: 123.0)

    msg: dict[str, object] = {
        "system": {
            "get_trouble": {
                "troubles": ["Low batt", "ac fail"],
                "at": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "error_code": E27ErrorCode.ELKERR_NONE,
            }
        }
    }
    assert handler(msg, _Ctx()) is True
    assert state.system_status.get("troubles") == ["Low batt", "ac fail"]


def test_system_get_troubles_variant_updates_snapshot() -> None:
    state = PanelState()
    emit = _EmitSpy()
    handler = make_system_get_trouble_handler(state, emit, now=lambda: 123.0)

    msg: dict[str, object] = {
        "system": {"get_troubles": {"troubles": [], "error_code": E27ErrorCode.ELKERR_NONE}}
    }
    assert handler(msg, _Ctx()) is True
    assert state.system_status.get("troubles") == []



"""
elke27_lib/features/system.py

Feature module: system
"""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from elke27_lib.kernel import E27Kernel

from elke27_lib.handlers.system import (
    make_system_file_info_handler,
    make_system_get_attribs_handler,
    make_system_get_cutoffs_handler,
    make_system_get_debug_flags_handler,
    make_system_get_debug_string_handler,
    make_system_get_sounders_handler,
    make_system_get_system_time_handler,
    make_system_get_table_info_handler,
    make_system_get_trouble_handler,
    make_system_get_troubles_handler,
    make_system_get_update_handler,
    make_system_r_u_alive_handler,
    make_system_reconfig_handler,
    make_system_reset_smokes_handler,
    make_system_set_attribs_handler,
    make_system_set_cutoffs_handler,
    make_system_set_debug_flags_handler,
    make_system_set_run_handler,
    make_system_set_system_key_handler,
    make_system_set_system_time_handler,
    make_system_start_updt_handler,
)

ROUTE_SYSTEM_GET_TROUBLE = ("system", "get_trouble")
ROUTE_SYSTEM_GET_TROUBLES = ("system", "get_troubles")
ROUTE_SYSTEM_GET_TABLE_INFO = ("system", "get_table_info")
ROUTE_SYSTEM_GET_ATTRIBS = ("system", "get_attribs")
ROUTE_SYSTEM_SET_ATTRIBS = ("system", "set_attribs")
ROUTE_SYSTEM_GET_CUTOFFS = ("system", "get_cutoffs")
ROUTE_SYSTEM_SET_CUTOFFS = ("system", "set_cutoffs")
ROUTE_SYSTEM_GET_SOUNDERS = ("system", "get_sounders")
ROUTE_SYSTEM_GET_SYSTEM_TIME = ("system", "get_system_time")
ROUTE_SYSTEM_SET_SYSTEM_TIME = ("system", "set_system_time")
ROUTE_SYSTEM_SET_SYSTEM_KEY = ("system", "set_system_key")
ROUTE_SYSTEM_FILE_INFO = ("system", "file_info")
ROUTE_SYSTEM_GET_DEBUG_FLAGS = ("system", "get_debug_flags")
ROUTE_SYSTEM_SET_DEBUG_FLAGS = ("system", "set_debug_flags")
ROUTE_SYSTEM_GET_DEBUG_STRING = ("system", "get_debug_string")
ROUTE_SYSTEM_R_U_ALIVE = ("system", "r_u_alive")
ROUTE_SYSTEM_RESET_SMOKES = ("system", "reset_smokes")
ROUTE_SYSTEM_SET_RUN = ("system", "set_run")
ROUTE_SYSTEM_START_UPDT = ("system", "start_updt")
ROUTE_SYSTEM_RECONFIG = ("system", "reconfig")
ROUTE_SYSTEM_GET_UPDATE = ("system", "get_update")


def register(elk: E27Kernel) -> None:
    elk.register_handler(
        ROUTE_SYSTEM_GET_TROUBLE,
        make_system_get_trouble_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_TROUBLES,
        make_system_get_troubles_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_TABLE_INFO,
        make_system_get_table_info_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_ATTRIBS,
        make_system_get_attribs_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_SET_ATTRIBS,
        make_system_set_attribs_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_CUTOFFS,
        make_system_get_cutoffs_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_SET_CUTOFFS,
        make_system_set_cutoffs_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_SOUNDERS,
        make_system_get_sounders_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_SYSTEM_TIME,
        make_system_get_system_time_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_SET_SYSTEM_TIME,
        make_system_set_system_time_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_SET_SYSTEM_KEY,
        make_system_set_system_key_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_FILE_INFO,
        make_system_file_info_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_DEBUG_FLAGS,
        make_system_get_debug_flags_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_SET_DEBUG_FLAGS,
        make_system_set_debug_flags_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_DEBUG_STRING,
        make_system_get_debug_string_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_R_U_ALIVE,
        make_system_r_u_alive_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_RESET_SMOKES,
        make_system_reset_smokes_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_SET_RUN,
        make_system_set_run_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_START_UPDT,
        make_system_start_updt_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_RECONFIG,
        make_system_reconfig_handler(elk.state, elk.emit, elk.now),
    )
    elk.register_handler(
        ROUTE_SYSTEM_GET_UPDATE,
        make_system_get_update_handler(elk.state, elk.emit, elk.now),
    )

    elk.register_request(
        ROUTE_SYSTEM_GET_TROUBLE,
        build_system_get_trouble_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_TROUBLES,
        build_system_get_trouble_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_TABLE_INFO,
        build_system_get_table_info_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_ATTRIBS,
        build_system_get_attribs_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_SET_ATTRIBS,
        build_system_set_attribs_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_CUTOFFS,
        build_system_get_cutoffs_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_SET_CUTOFFS,
        build_system_set_cutoffs_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_SOUNDERS,
        build_system_get_sounders_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_SYSTEM_TIME,
        build_system_get_system_time_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_SET_SYSTEM_TIME,
        build_system_set_system_time_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_SET_SYSTEM_KEY,
        build_system_set_system_key_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_FILE_INFO,
        build_system_file_info_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_DEBUG_FLAGS,
        build_system_get_debug_flags_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_SET_DEBUG_FLAGS,
        build_system_set_debug_flags_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_DEBUG_STRING,
        build_system_get_debug_string_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_R_U_ALIVE,
        build_system_r_u_alive_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_RESET_SMOKES,
        build_system_reset_smokes_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_SET_RUN,
        build_system_set_run_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_START_UPDT,
        build_system_start_updt_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_RECONFIG,
        build_system_reconfig_payload,
    )
    elk.register_request(
        ROUTE_SYSTEM_GET_UPDATE,
        build_system_get_update_payload,
    )


def build_system_get_trouble_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_get_table_info_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_get_attribs_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_set_attribs_payload(**kwargs: Any) -> Mapping[str, Any]:
    if not kwargs:
        raise ValueError("build_system_set_attribs_payload requires at least one attribute.")
    return dict(kwargs)


def build_system_get_cutoffs_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_set_cutoffs_payload(**kwargs: Any) -> Mapping[str, Any]:
    if not kwargs:
        raise ValueError("build_system_set_cutoffs_payload requires at least one cutoff value.")
    return dict(kwargs)


def build_system_get_sounders_payload(
    *, sounder_id: int | None = None, **_kwargs: Any
) -> Mapping[str, Any]:
    payload: dict[str, object] = {}
    if sounder_id is not None:
        if sounder_id < 0:
            raise ValueError(
                f"build_system_get_sounders_payload: sounder_id must be int >= 0 (got {sounder_id!r})"
            )
        payload["sounder_id"] = sounder_id
    return payload


def build_system_get_system_time_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_set_system_time_payload(
    *,
    tz_offset: int,
    city_index: int,
    gmt_seconds: int,
    dst_active: bool,
    **_kwargs: Any,
) -> Mapping[str, Any]:
    if city_index < 0:
        raise ValueError("build_system_set_system_time_payload: city_index must be int >= 0")
    if gmt_seconds < 0:
        raise ValueError("build_system_set_system_time_payload: gmt_seconds must be int >= 0")
    return {
        "tz_offset": tz_offset,
        "city_index": city_index,
        "gmt_seconds": gmt_seconds,
        "dst_active": dst_active,
    }


def build_system_set_system_key_payload(*, key: int, **_kwargs: Any) -> Mapping[str, Any]:
    if key < 0:
        raise ValueError("build_system_set_system_key_payload: key must be int >= 0")
    return {"key": key}


def build_system_file_info_payload(
    *,
    file_list: bool | None = None,
    file_num: int | None = None,
    **_kwargs: Any,
) -> Mapping[str, Any]:
    payload: dict[str, object] = {}
    if file_list is not None:
        payload["file_list"] = file_list
    if file_num is not None:
        if file_num < 0:
            raise ValueError("build_system_file_info_payload: file_num must be int >= 0")
        payload["file_num"] = file_num
    if not payload:
        raise ValueError("build_system_file_info_payload requires file_list or file_num")
    return payload


def build_system_get_debug_flags_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_set_debug_flags_payload(
    *,
    dbug: list[int] | None = None,
    dbug_id: int | None = None,
    dbug_not_id: int | None = None,
    **_kwargs: Any,
) -> Mapping[str, Any]:
    payload: dict[str, object] = {}
    if dbug is not None:
        payload["dbug"] = dbug
    if dbug_id is not None:
        if dbug_id < 0:
            raise ValueError("build_system_set_debug_flags_payload: dbug_id must be int >= 0")
        payload["dbug_id"] = dbug_id
    if dbug_not_id is not None:
        if dbug_not_id < 0:
            raise ValueError("build_system_set_debug_flags_payload: dbug_not_id must be int >= 0")
        payload["dbug_not_id"] = dbug_not_id
    if not payload:
        raise ValueError("build_system_set_debug_flags_payload requires at least one field.")
    return payload


def build_system_get_debug_string_payload(*, dbug_id: int, **_kwargs: Any) -> Mapping[str, Any]:
    if dbug_id < 0:
        raise ValueError("build_system_get_debug_string_payload: dbug_id must be int >= 0")
    return {"dbug_id": dbug_id}


def build_system_r_u_alive_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_reset_smokes_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {"reset_smokes": True}


def build_system_set_run_payload(*, app: str, **_kwargs: Any) -> Mapping[str, Any]:
    if not app.strip():
        raise ValueError("build_system_set_run_payload: app must be a non-empty string")
    return {"app": app}


def build_system_start_updt_payload(
    *, device_id: str, ft: int, **_kwargs: Any
) -> Mapping[str, Any]:
    if not device_id.strip():
        raise ValueError("build_system_start_updt_payload: device_id must be a non-empty string")
    if ft < 0:
        raise ValueError("build_system_start_updt_payload: ft must be int >= 0")
    return {"device_id": device_id, "ft": ft}


def build_system_reconfig_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}


def build_system_get_update_payload(**_kwargs: Any) -> Mapping[str, Any]:
    return {}



from __future__ import annotations

import json
from datetime import UTC, datetime

import pytest

from elke27_lib import (
    ClientConfig,
    Elke27Client,
    Elke27Event,
    EventType,
    PanelInfo,
    PanelSnapshot,
    TableInfo,
)
from elke27_lib.dispatcher import DispatchContext, MessageKind
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    AreaStatusUpdated,
    CsmSnapshotUpdated,
    DomainCsmChanged,
    Event,
    TableCsmChanged,
)
from elke27_lib.handlers.zone import (
    make_zone_get_attribs_handler,
    make_zone_get_defs_handler,
)
from elke27_lib.types import CsmSnapshot


def _empty_payload(**_kwargs: object) -> dict[str, object]:
    return {}


def _make_event(seq: int) -> Elke27Event:
    return Elke27Event(
        event_type=EventType.SYSTEM,
        data={"seq": seq},
        seq=seq,
        timestamp=datetime.now(UTC),
        raw_type="test",
    )


def _ctx(route: tuple[str, str]) -> DispatchContext:
    return DispatchContext(
        kind=MessageKind.DIRECTED,
        seq=None,
        session_id=None,
        route=route,
        classification="RESPONSE",
    )


def test_snapshot_atomic_replacement_and_version() -> None:
    client = Elke27Client()
    initial = client.snapshot

    client._replace_snapshot(panel_info=PanelInfo(model="M1"), table_info=TableInfo())
    snap1 = client.snapshot
    client._replace_snapshot(panel_info=PanelInfo(model="M2"))
    snap2 = client.snapshot

    assert isinstance(initial, PanelSnapshot)
    assert initial.version == 0
    assert initial.zone_definitions == {}
    assert initial.output_definitions == {}
    assert snap1.version == 1
    assert snap2.version == 2
    assert initial.panel.model is None
    assert snap1.panel.model == "M1"
    assert snap2.panel.model == "M2"


def test_event_queue_drop_oldest() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    evt1 = _make_event(1)
    evt2 = _make_event(2)
    evt3 = _make_event(3)

    client._enqueue_event(evt1)
    client._enqueue_event(evt2)
    client._enqueue_event(evt3)

    assert client._event_queue.qsize() == 2
    first = client._event_queue.get_nowait()
    second = client._event_queue.get_nowait()
    assert first.seq == 2
    assert second.seq == 3


@pytest.mark.asyncio
async def test_events_iterator_terminates_on_disconnect() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    evt = _make_event(1)
    client._enqueue_event(evt)
    client._signal_event_stream_end()

    seen: list[Elke27Event] = []
    async for item in client.events():
        seen.append(item)

    assert len(seen) == 1
    assert seen[0].seq == 1


def test_subscriber_callback_exception_does_not_break_queue() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))

    def _bad_callback(evt: Elke27Event) -> None:
        del evt
        raise RuntimeError("boom")

    client.subscribe(_bad_callback)

    evt = AreaStatusUpdated(
        kind=AreaStatusUpdated.KIND,
        at=UNSET_AT,
        seq=UNSET_SEQ,
        classification=UNSET_CLASSIFICATION,
        route=UNSET_ROUTE,
        session_id=UNSET_SESSION_ID,
        area_id=1,
        changed_fields=(),
    )
    client._handle_kernel_event(evt)

    queued = client._event_queue.get_nowait()
    assert isinstance(queued, Elke27Event)
    json.dumps(queued.data)


def test_typed_subscriber_receives_event() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    seen: list[Event] = []

    def _typed_cb(evt: Event) -> None:
        seen.append(evt)

    client.subscribe_typed(_typed_cb)

    evt = AreaStatusUpdated(
        kind=AreaStatusUpdated.KIND,
        at=UNSET_AT,
        seq=UNSET_SEQ,
        classification=UNSET_CLASSIFICATION,
        route=UNSET_ROUTE,
        session_id=UNSET_SESSION_ID,
        area_id=1,
        changed_fields=(),
    )
    client._handle_kernel_event(evt)
    assert len(seen) == 1
    assert isinstance(seen[0], AreaStatusUpdated)


def test_typed_subscriber_receives_csm_events() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    seen: list[Event] = []

    def _typed_cb(evt: Event) -> None:
        seen.append(evt)

    client.subscribe_typed(_typed_cb)
    snapshot = CsmSnapshot(
        domain_csms={"zone": 1},
        table_csms={"zone": 2},
        version=1,
        updated_at=datetime.now(UTC),
    )

    events = [
        DomainCsmChanged(
            kind=DomainCsmChanged.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            csm_domain="zone",
            old=None,
            new=1,
        ),
        TableCsmChanged(
            kind=TableCsmChanged.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            csm_domain="zone",
            old=None,
            new=2,
        ),
        CsmSnapshotUpdated(
            kind=CsmSnapshotUpdated.KIND,
            at=UNSET_AT,
            seq=UNSET_SEQ,
            classification=UNSET_CLASSIFICATION,
            route=UNSET_ROUTE,
            session_id=UNSET_SESSION_ID,
            snapshot=snapshot,
        ),
    ]
    for evt in events:
        client._handle_kernel_event(evt)

    assert [type(item) for item in seen] == [type(evt) for evt in events]


def test_area_bypass_count_zero_does_not_refresh_bypassed_zones() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    kernel = client._kernel
    state = kernel.state
    area = state.get_or_create_area(1)
    area.num_bypassed_zones = 0
    zone = state.get_or_create_zone(33)
    zone.area_id = 1
    zone.bypassed = True

    recorded: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _fake_request(route: tuple[str, str], **kwargs: object) -> int:
        recorded.append((route, dict(kwargs)))
        return 1

    kernel.request = _fake_request
    kernel.requests.register(("zone", "get_all_zones_status"), _empty_payload)

    evt = AreaStatusUpdated(
        kind=AreaStatusUpdated.KIND,
        at=UNSET_AT,
        seq=UNSET_SEQ,
        classification=UNSET_CLASSIFICATION,
        route=UNSET_ROUTE,
        session_id=UNSET_SESSION_ID,
        area_id=1,
        changed_fields=("num_bypassed_zones",),
    )
    client._handle_kernel_event(evt)

    assert recorded == []


def test_area_bypass_count_nonzero_does_not_refresh_non_bypassed_zones() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    kernel = client._kernel
    state = kernel.state
    area = state.get_or_create_area(1)
    area.num_bypassed_zones = 1
    zone_bypassed = state.get_or_create_zone(33)
    zone_bypassed.area_id = 1
    zone_bypassed.bypassed = True
    zone_open = state.get_or_create_zone(34)
    zone_open.area_id = 1
    zone_open.bypassed = False

    recorded: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _fake_request(route: tuple[str, str], **kwargs: object) -> int:
        recorded.append((route, dict(kwargs)))
        return 1

    kernel.request = _fake_request
    kernel.requests.register(("zone", "get_all_zones_status"), _empty_payload)

    evt = AreaStatusUpdated(
        kind=AreaStatusUpdated.KIND,
        at=UNSET_AT,
        seq=UNSET_SEQ,
        classification=UNSET_CLASSIFICATION,
        route=UNSET_ROUTE,
        session_id=UNSET_SESSION_ID,
        area_id=1,
        changed_fields=("num_bypassed_zones",),
    )
    client._handle_kernel_event(evt)

    assert recorded == []


def test_area_bypass_refresh_suppressed_after_local_command() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    kernel = client._kernel
    state = kernel.state
    area = state.get_or_create_area(1)
    area.num_bypassed_zones = 1
    zone = state.get_or_create_zone(33)
    zone.area_id = 1

    recorded: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _fake_request(route: tuple[str, str], **kwargs: object) -> int:
        recorded.append((route, dict(kwargs)))
        return 1

    kernel.request = _fake_request
    kernel.requests.register(("zone", "get_all_zones_status"), _empty_payload)

    def _fake_now() -> float:
        return 100.0

    kernel.now = _fake_now
    client._record_local_zone_bypass(33)

    evt = AreaStatusUpdated(
        kind=AreaStatusUpdated.KIND,
        at=UNSET_AT,
        seq=UNSET_SEQ,
        classification=UNSET_CLASSIFICATION,
        route=UNSET_ROUTE,
        session_id=UNSET_SESSION_ID,
        area_id=1,
        changed_fields=("num_bypassed_zones",),
    )
    client._handle_kernel_event(evt)

    assert recorded == []


def test_area_status_no_change_refreshes_bulk_status() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    kernel = client._kernel
    recorded: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _fake_request(route: tuple[str, str], **kwargs: object) -> int:
        recorded.append((route, dict(kwargs)))
        return 1

    kernel.request = _fake_request
    kernel.requests.register(("zone", "get_all_zones_status"), _empty_payload)

    evt = AreaStatusUpdated(
        kind=AreaStatusUpdated.KIND,
        at=UNSET_AT,
        seq=UNSET_SEQ,
        classification=UNSET_CLASSIFICATION,
        route=UNSET_ROUTE,
        session_id=UNSET_SESSION_ID,
        area_id=1,
        changed_fields=(),
    )
    client._handle_kernel_event(evt)

    assert (("zone", "get_all_zones_status"), {}) in recorded


def test_snapshot_includes_zone_definitions() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    state = client._kernel.state
    zone = state.get_or_create_zone(1)
    zone.name = "Front Door"
    zone.definition = 1
    zone.attribs["zone_type"] = "Window"
    zone.attribs["kind"] = "Door"
    state.zone_defs_by_id[1] = {"definition": "BURG PERIM INST"}

    client._replace_snapshot(
        panel_info=client._build_panel_info(),
        table_info=client._build_table_info(),
        areas=client._build_area_map(),
        zones=client._build_zone_map(),
        zone_definitions=client._build_zone_definitions(),
        outputs=client._build_output_map(),
        output_definitions=client._build_output_definitions(),
    )
    snapshot = client.snapshot
    zone_def = snapshot.zone_definitions[1]
    assert zone_def.name == "Front Door"
    assert zone_def.definition == "BURG PERIM INST"
    assert zone_def.zone_type == "Window"
    assert zone_def.kind == "Door"


def test_zone_attribs_handler_updates_snapshot_definitions() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    state = client._kernel.state
    emitted: list[Event] = []

    def _emit(evt: Event, _ctx: DispatchContext) -> None:
        emitted.append(evt)

    handler = make_zone_get_attribs_handler(state, _emit, now=lambda: 123.0)

    msg = {
        "zone": {
            "get_attribs": {
                "zone_id": 1,
                "name": "Front",
                "definition": "BURG PERIM INST",
                "zone_type": "Window",
            }
        }
    }
    assert handler(msg, _ctx(("zone", "get_attribs"))) is True
    for evt in emitted:
        client._handle_kernel_event(evt)

    zone_def = client.snapshot.zone_definitions[1]
    assert zone_def.name == "Front"
    assert zone_def.definition == "BURG PERIM INST"
    assert zone_def.zone_type == "Window"


def test_refresh_zone_config_requests_definitions_and_attribs() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    kernel = client._kernel
    state = kernel.state
    state.inventory.configured_zones = {1, 2}
    recorded: list[tuple[tuple[str, str], dict[str, object]]] = []

    def _fake_request(route: tuple[str, str], **kwargs: object) -> int:
        recorded.append((route, dict(kwargs)))
        return 1

    kernel.request = _fake_request
    kernel.requests.register(("zone", "get_table_info"), _empty_payload)
    kernel.requests.register(("zone", "get_configured"), _empty_payload)
    kernel.requests.register(("zone", "get_defs"), _empty_payload)
    kernel.requests.register(("zone", "get_attribs"), _empty_payload)
    client._refresh_zone_config()

    assert ("zone", "get_table_info") in [route for route, _ in recorded]
    assert ("zone", "get_configured") in [route for route, _ in recorded]
    assert ("zone", "get_defs") in [route for route, _ in recorded]
    attrib_routes = [route for route, _ in recorded if route == ("zone", "get_attribs")]
    assert len(attrib_routes) == 2


def test_bootstrap_zone_definitions_not_none() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    state = client._kernel.state
    emitted: list[Event] = []

    def _emit(evt: Event, _ctx: DispatchContext) -> None:
        emitted.append(evt)

    defs_handler = make_zone_get_defs_handler(state, _emit, now=lambda: 123.0)
    attribs_handler = make_zone_get_attribs_handler(state, _emit, now=lambda: 123.0)

    defs_msg = {"zone": {"get_defs": {"definitions": ["BURG PERIM INST"], "block_id": 1}}}
    attribs_msg = {
        "zone": {
            "get_attribs": {
                "zone_id": 1,
                "name": "Front",
                "definition": "BURG PERIM INST",
            }
        }
    }
    assert defs_handler(defs_msg, _ctx(("zone", "get_defs"))) is True
    assert attribs_handler(attribs_msg, _ctx(("zone", "get_attribs"))) is True
    for evt in emitted:
        client._handle_kernel_event(evt)

    zone_def = client.snapshot.zone_definitions[1]
    assert zone_def.definition is not None


def test_refresh_zone_config_updates_snapshot_after_handlers() -> None:
    client = Elke27Client(config=ClientConfig(event_queue_size=2))
    state = client._kernel.state
    emitted: list[Event] = []

    def _emit(evt: Event, _ctx: DispatchContext) -> None:
        emitted.append(evt)

    defs_handler = make_zone_get_defs_handler(state, _emit, now=lambda: 123.0)
    attribs_handler = make_zone_get_attribs_handler(state, _emit, now=lambda: 123.0)

    defs_msg = {"zone": {"get_defs": {"definitions": ["BURG PERIM INST"], "block_id": 1}}}
    attribs_msg = {
        "zone": {
            "get_attribs": {
                "zone_id": 1,
                "name": "Front",
                "definition": "BURG PERIM INST",
                "zone_type": "Window",
            }
        }
    }
    assert defs_handler(defs_msg, _ctx(("zone", "get_defs"))) is True
    assert attribs_handler(attribs_msg, _ctx(("zone", "get_attribs"))) is True
    for evt in emitted:
        client._handle_kernel_event(evt)

    zone_def = client.snapshot.zone_definitions[1]
    assert zone_def.name == "Front"
    assert zone_def.definition == "BURG PERIM INST"
    assert zone_def.zone_type == "Window"



from __future__ import annotations

from typing import Any


def get_private(obj: object, name: str) -> Any:
    return getattr(obj, name)


def set_private(obj: object, name: str, value: Any) -> None:
    setattr(obj, name, value)


def get_kernel(client: object) -> Any:
    return get_private(client, "_kernel")



import asyncio
import time
from collections.abc import Callable

import pytest

from elke27_lib.outbound import OutboundItem, OutboundPriority, OutboundQueue


@pytest.mark.asyncio
async def test_outbound_queue_throttle() -> None:
    loop = asyncio.get_running_loop()
    sent_times: list[float] = []
    done = asyncio.Event()

    def _send(payload: bytes) -> None:
        del payload

    def _on_sent(ts: float) -> None:
        sent_times.append(ts)
        if len(sent_times) >= 5:
            done.set()

    queue = OutboundQueue(loop=loop, send_fn=_send, min_interval_s=0.05, max_burst=1)
    queue.start()

    for i in range(5):
        queue.enqueue(
            OutboundItem(
                payload=b"x",
                seq=i,
                kind="request",
                priority=OutboundPriority.NORMAL,
                enqueued_at=time.monotonic(),
                on_sent=_on_sent,
            )
        )

    await asyncio.wait_for(done.wait(), timeout=2.0)
    deltas = [b - a for a, b in zip(sent_times, sent_times[1:], strict=False)]
    assert all(delta >= 0.04 for delta in deltas)


@pytest.mark.asyncio
async def test_outbound_queue_priority() -> None:
    loop = asyncio.get_running_loop()
    sent: list[str] = []
    done = asyncio.Event()

    def _send(payload: bytes) -> None:
        del payload

    def _make_on_sent(label: str) -> Callable[[float], None]:
        def _cb(_: float) -> None:
            sent.append(label)
            if len(sent) >= 4:
                done.set()

        return _cb

    queue = OutboundQueue(loop=loop, send_fn=_send, min_interval_s=0.01, max_burst=1)
    queue.start()

    queue.enqueue(
        OutboundItem(
            payload=b"n1",
            seq=1,
            kind="request",
            priority=OutboundPriority.NORMAL,
            enqueued_at=time.monotonic(),
            on_sent=_make_on_sent("n1"),
        )
    )
    queue.enqueue(
        OutboundItem(
            payload=b"n2",
            seq=2,
            kind="request",
            priority=OutboundPriority.NORMAL,
            enqueued_at=time.monotonic(),
            on_sent=_make_on_sent("n2"),
        )
    )
    queue.enqueue(
        OutboundItem(
            payload=b"n3",
            seq=3,
            kind="request",
            priority=OutboundPriority.NORMAL,
            enqueued_at=time.monotonic(),
            on_sent=_make_on_sent("n3"),
        )
    )
    queue.enqueue(
        OutboundItem(
            payload=b"h1",
            seq=99,
            kind="request",
            priority=OutboundPriority.HIGH,
            enqueued_at=time.monotonic(),
            on_sent=_make_on_sent("h1"),
        )
    )

    await asyncio.wait_for(done.wait(), timeout=2.0)
    assert "h1" in sent
    assert sent.index("h1") < sent.index("n3")


@pytest.mark.asyncio
async def test_outbound_queue_stop_fails_pending() -> None:
    loop = asyncio.get_running_loop()
    failed = asyncio.Event()

    def _send(payload: bytes) -> None:
        del payload

    def _on_fail(exc: BaseException) -> None:
        assert isinstance(exc, RuntimeError)
        failed.set()

    queue = OutboundQueue(loop=loop, send_fn=_send, min_interval_s=0.5, max_burst=1)
    queue.start()

    queue.enqueue(
        OutboundItem(
            payload=b"x",
            seq=1,
            kind="request",
            priority=OutboundPriority.NORMAL,
            enqueued_at=time.monotonic(),
            on_fail=_on_fail,
        )
    )

    queue.stop(fail_exc=RuntimeError("transport gone"))
    await asyncio.wait_for(failed.wait(), timeout=1.0)



"""
elke27_lib/handlers/rule.py

Read-only handlers for the "rule" domain.
"""

from __future__ import annotations

from collections.abc import Callable, Mapping
from typing import Any, cast

from elke27_lib.dispatcher import DispatchContext
from elke27_lib.events import (
    UNSET_AT,
    UNSET_CLASSIFICATION,
    UNSET_ROUTE,
    UNSET_SEQ,
    UNSET_SESSION_ID,
    ApiError,
    AuthorizationRequiredEvent,
    Event,
)
from elke27_lib.states import PanelState

EmitFn = Callable[[Event, DispatchContext], None]
NowFn = Callable[[], float]


def _as_mapping(obj: object) -> Mapping[str, Any] | None:
    if isinstance(obj, Mapping):
        return cast(Mapping[str, Any], obj)
    return None


def make_rule_get_rules_handler(state: PanelState, emit: EmitFn, now: NowFn):
    """
    Handler for ("rule","get_rules").
    """

    def handler_rule_get_rules(msg: Mapping[str, Any], ctx: DispatchContext) -> bool:
        rule_obj = _as_mapping(msg.get("rule"))
        if rule_obj is None:
            return False

        payload = _as_mapping(rule_obj.get("get_rules"))
        if payload is None:
            return False

        error_code = payload.get("error_code", rule_obj.get("error_code"))
        if isinstance(error_code, int) and error_code != 0:
            if error_code == 11008:
                emit(
                    AuthorizationRequiredEvent(
                        kind=AuthorizationRequiredEvent.KIND,
                        at=UNSET_AT,
                        seq=UNSET_SEQ,
                        classification=UNSET_CLASSIFICATION,
                        route=UNSET_ROUTE,
                        session_id=UNSET_SESSION_ID,
                        error_code=error_code,
                        scope="rule",
                        entity_id=None,
                        message=None,
                    ),
                    ctx,
                )
                return True
            emit(
                ApiError(
                    kind=ApiError.KIND,
                    at=UNSET_AT,
                    seq=UNSET_SEQ,
                    classification=UNSET_CLASSIFICATION,
                    route=UNSET_ROUTE,
                    session_id=UNSET_SESSION_ID,
                    error_code=error_code,
                    scope="rule",
                    entity_id=None,
                    message=None,
                ),
                ctx,
            )
            return True

        block_id = payload.get("block_id")
        if not isinstance(block_id, int) or block_id < 0:
            return False

        block_count = payload.get("block_count")
        if isinstance(block_count, int):
            state.rules_block_count = block_count

        if block_id == 0:
            state.rules = {}
        data = payload.get("data")
        if isinstance(data, str) and block_id >= 1:
            state.rules[block_id] = {"block_id": block_id, "data": data}

        state.panel.last_message_at = now()
        return True

    return handler_rule_get_rules
